#include <cxxtls/GaussianLpSolver.h>
#include <cxxtls/strtool.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/foreach.h>
#include <vector>
#include <iostream>
#include <stdlib.h>
#include <math.h>

using namespace std;
using namespace cxxtls;


int main(int argc, char **argv)
{
   vector<StrTool::stringlist_t> input;

   string line;

   if(argc > 1)
   {
      cout << "Error:  this program, solve.exe, takes no parameters.\n";
      cout << "The program accepts N row of N+1 columns of floating point numbers\n";
      cout << "from stdin.\n\n";
      cout << "It uses the Gaussian method of solving N equations and N knowns.\n\n";
      cout << "The number of lines in the file defines N -- the number equations.\n";
      cout << "and the number of unknowns.\n\n";
      cout << "There must be N+1 floats on each line of input.  The first N are the parameters\n";
      cout << "of each equation.  The last floating point number on each line is the right\n";
      cout << "hand side of the equation.  If a solution was found, program will output\n";
      cout << "the value of the solution found, all N of them, in the follwing format:\n";
      cout << "---\n";
      cout << "    value[1] value[2] ... value[N]\n";
      cout << "Otherwise, an error message will be presented\n";
      exit(3);

   }


   while(std::getline(cin, line) )
   {
      input.push_back(StrTool::stringlist_t());
      
      StrTool::parse_words(line, &input.back(), 10000, StrTool::Is_Space());
   }

   size_t inputLines = input.size();

   size_t row;

   for(row = 0; row != inputLines; ++row)
   {
      if(input[row].size() != 1 + inputLines)
      {
         cout << "Error:  row " << row+1 << " contains " << input[row].size() << " elements instead of the required " << 1 + inputLines << endl;
         cout << "        row contents are as follows\n";
         CXXTLS_FOREACH(std::string const & value, input[row])
         {
            cout << "\t'" << value << "'\n";
         }
         exit(1);
      }
   }

   // if we get here, we know that the right number of equations and parameters have been supplied -- but they might not be properly formatted.

   GaussianLpSolver solver(inputLines);

   for(row = 0; row < inputLines; ++row)
   {
      size_t col;

      double d;

      StrTool::stringlist_t::iterator fcol = input[row].begin();

      for(col = 0; col < inputLines ; ++col, ++fcol)
      {
          d = lexical_cast<double>(*fcol);
          
          solver.setParameter(col, row, d);
      }

      d = lexical_cast<double>(*fcol);

      solver.setRHS(row, d);

   }

   if(solver.solve())
   {
      for(row = 0; row < inputLines; ++row)
      {
         if(isnan(solver.result(row)))
         {
            cout << "Error:  could not solve" << endl;
            exit(2);
         }
      }

      cout << "---\n";

      for(row = 0; row < inputLines; ++row)
      {
          cout << solver.result(row) << ' ';
      }

      cout << endl;
   }
   else
   {
      cout << "Error:  could not solve" << endl;
      exit(1);
   }

   exit(0);



}
