#include <BrashCommands.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/stringexpressionparser.h>


using namespace std;
using namespace cxxtls;

static string colon(":");
static string nl("\n");
static string comma(",");
   


inline bool hasBrashContinuationFlag(ScriptInputLine const &line)
{
    if(    !line.text_.empty()
       &&   line.text_[line.text_.size()-1] == '\n'
       &&   line.text_[line.text_.size()-2] == '\1'
      )
    {
       return true;
    }

    return false;
}


static bool parseContinuedLine(ScriptLineStream  &stream,
                               ScriptInputLine   &command,
                               bool               pushPromptDepth=false
                              )
    /// read a single command line from the input stream and handle
    /// line continuations
{
    if(stream.readLine(command))
    {
        while( hasBrashContinuationFlag(command) )
        {
            size_t trailingBackSlash = command.text_.find_last_of('\\');

            command.text_.erase(trailingBackSlash, command.text_.size());

            bool foundMore = false;

            ScriptInputLine more;

            bool foundMoreLines = stream.readLine(more);

            if(!foundMoreLines)
            {
               command.text_ += '\n';  // put the trailing \n back on teh line
               break;
            }

            command.text_ += more.text_;  

        }

        return true;
    }

    return false;
}

string parseFirstWord(string::const_iterator &f, string::const_iterator const &l)
{
   StringExpr::skipSeparators(f,l);

   string rv;

   while(f !=l && isalnum(*f))
   {
      rv += *f++;
   }

   if(f == l)
      return rv;

   static string emptystring;

   if(!isspace(*f))
      return emptystring;

   return rv;
    
}

static bool parseCommand  (string::const_iterator           &f, 
                           string::const_iterator           &l,
                           ScriptLineStream                 &stream,
                           BrashStatementList               &output,
                           ScriptThread::ManagedFileHandles &messageFiles,
                           ScriptInputLine                  &command   // over which f and l iterate
                          )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseIfStatement( string::const_iterator           &f, 
                              string::const_iterator           &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,  
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}


static bool parseWhileStatement( string::const_iterator           &f, 
                                 string::const_iterator           &l,
                                 ScriptLineStream                 &stream,
                                 BrashStatementList               &output,
                                 ScriptThread::ManagedFileHandles &messageFiles,
                                 ScriptInputLine                  &command   // over which f and l iterate
                               )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseForStatement( string::const_iterator           &f, 
                               string::const_iterator           &l,
                               ScriptLineStream                 &stream,
                               BrashStatementList               &output,
                               ScriptThread::ManagedFileHandles &messageFiles,  
                               ScriptInputLine                  &command   // over which f and l iterate
                             )
{
   
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseReturnStatement( string::const_iterator       &f, 
                                  string::const_iterator       &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseBreakStatement( string::const_iterator        &f, 
                                 string::const_iterator        &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseContinueStatement( string::const_iterator     &f, 
                                    string::const_iterator     &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}


static bool parseExitStatement( string::const_iterator         &f, 
                                string::const_iterator         &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,  
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseAliasStatement( string::const_iterator        &f, 
                                 string::const_iterator        &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}

static bool parseLetStatement( string::const_iterator          &f, 
                               string::const_iterator          &l,
                              ScriptLineStream                 &stream,
                              BrashStatementList               &output,
                              ScriptThread::ManagedFileHandles &messageFiles,
                              ScriptInputLine                  &command   // over which f and l iterate
                            )
{
   // f and l iterate over command
   messageFiles.writeBytes(2, string("Error:  unhandled statement:  ") + command.text_ + nl  );
   return false;
}


static bool parseStatement(ScriptLineStream                 &stream,
                           BrashStatementList               &output,
                           ScriptThread::ManagedFileHandles &messageFiles  
                          )
    ///  parse lines from stream that comprise a single bash statement
{
    ScriptInputLine command;     // individual parses may refil as the while loop below progresses.

    if(!parseContinuedLine(stream, command))
       return false;  // do not ignore i/o errors

    string::const_iterator f = command.text_.begin(),
                           l = command.text_.end();

    StringExpr::skipSeparators(f,l);


    while(f != l)
    {   // parse all commands on the line -- handling continuations in the command-
        // sub-type parsers
    
        string firstWord = parseFirstWord(f, l);
    
        if(firstWord.empty())
        {
           return true; // just ignore empty lines
        }
    
        // if we get here, the current line has at least one word one it.
    
        char fc = 0;
    
        if(f != l)
           fc = *f;
    
    
        f = command.text_.begin();  // set up to restart the parsing process

        if(fc == '#')
           return true;  // just ignore comments
    
        if(fc  == '=' && !firstWord.empty())
        {
           // we either have an assignment statement or we have a set of variable superseeds
           // before command line statement.
    
           return parseCommand(f,l, stream, output, messageFiles, command);
    
        }
        else
        if(fc== 'i' && firstWord == "if")
        {
           return parseIfStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'w' && firstWord == "while")
        {
           return parseWhileStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'f' && firstWord == "for")
        {
           return parseForStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'r' && firstWord == "return")
        {
           return parseForStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'b' && firstWord == "break")
        {
           return parseBreakStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'c' && firstWord == "continue")
        {
           return parseContinueStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'e' && firstWord == "exit")
        {
           return parseExitStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'a' && firstWord == "alias")
        {
           return parseAliasStatement(f,l, stream, output, messageFiles, command);
        }
        else
        if(fc == 'l' && firstWord == "let")
        {
           return parseLetStatement(f,l, stream, output, messageFiles, command);
        }
        else
        {
            return parseCommand(f,l, stream, output, messageFiles, command);
        }

        StringExpr::skipSeparators(f,l);

    } // parse all commands on the line

    return true;
}



bool parseBrash(ScriptLineStream                 &stream,
                BrashStatementList               &output,
                ScriptThread::ManagedFileHandles &messageFiles  
               )
{
try
{
     while(!stream.eof())
     {
        if(!parseStatement(stream, output, messageFiles))
           return false;
     }

     return true;

} catch(std::exception const &e)
{

    messageFiles.writeBytes(2, string("Unhandled exception  ") + string(e.what()) + comma + string("while parsing ") + stream.filename() + colon + lexical_cast<string>(stream.lineNumber()) + nl );

    return false;
}
}


