#!/bin/bash
#outputExtension=.lnktree

# The outputExtension comment above must be preserved exactly as it is and on line2 of this script!

# This script expects 2 parameters:  the input file and the output file.
# The output file must have the file name extensions specified above.
#
# This script creates the output file
#
# The input file expected to be a linux object module, library, shared object, or executable.
# The purpose of this script is to produce a .tagpp file corresponding to the data found in the
# input file.  When objects are compiled for debug then full filename and line number information
# is available but if not, then no information will be be available and the symbols are assumed
# to be defined in some unknown loacation.

if [ $# != 2 ]
then
    exit 1
fi

rm -fr "$2"  # remove the output

exec >$2 2>&1  # redirect all stdout ouput to $2

echo "$1"
echo "DEFINED SYMBOLS"


#
#  Get the file name extension of the input
#
ext="${1##*.}"

#
#  see if there are any source files in this directory that could have created the object module
#

base="${1%.$ext}"

possibility=`ls $base.* | egrep -v "$base.o|$base.so|$base.a|$base.bin|$base.exe" | head`

if [ "$possibility" = "" ]
then
   possibility="`pwd`/unknown.txt"
fi

possibility="$possibility:1"

#
#  symbol possibility is now set up to contain a possible source for any symbols found in
#  the object module we are working on.  Note that the object module may well define the
#  correct source location, but if it doesn't, we are prepared with a default.
#

tmp=/tmp/$(basename $1).$$

nm -C -g -l "$1" | sed -e \
"
    s/^[0-9a-fA-F]\+ //1

    y/\t/ /

    s/^  *//1


    /^V /d   # remove virtual table symbol lines

    /^u/d    # some wierd helper symbols

    /^U __gxx_personality_v/d
    /^U vtable for/d

    /: *\$/d  # get rid of object module header lines


    #
    #  translate symbol type codes into human readable form
    #

    s/^[Uu] /EXTERNAL /1
    s/^[Tt] /DEFINED  /1
    s/^[Ww] /INLINED  /1
    s/^[BbDd] /DATA     /1


    #
    #  convert 'operator /' to 'operator div'
    #

    s%operator */%operator div%1


    s/\([^ ]\+\):0 *\$//1

    /:[0-9]\+ *\$/!{
		      # when a symbol line does not contain a filename, use the possibility we created
		      # above

		      s%.*%,&%1

		      /, *\$/d  # delete malformed lines

		      p
		      d


		  }


    s/^\(.*\) \(\/[^:]*:[0-9]\+\)$/\2,\1/1     # move the filename:linenumber to the front and put , on it

    /, *\$/d  # delete malformed lines


"  | sort -u

