#include <iostream>
#include <stdlib.h>
#include <cxxtls/charset.h>
#include <memory>
#include <string.h>

using namespace std;
using namespace cxxtls;

void bomb1(char const *msg, char const *file, int line)  { cerr << file << ':' << line << ":     " << msg << endl ; exit(1); }

#define bomb(msg)  bomb1(msg, __FILE__, __LINE__)

#define ok(msg)    cout << "    " << msg << " test passed" << endl;

int main()
{
    {
         cout << "Empty Charset test" << endl;

         Charset  set;

         for(int i = 0; i < 256; ++i)
             if(set.test(i))
                 bomb("empty charset had a mistaken bit set");

         ok("Empty Charset Test");

    }

    {
         cout << "Full Charset test" << endl;

         Charset  set;

         for(int i = 0; i < 256; ++i)
            set.set(i,true);


         for(int i = 0; i < 256; ++i)
             if(!set.test(i))
                 bomb("full charset had a mistaken bit set");

         ok("Full Charset Test");

    }

    {
         cout << "One character Charset test" << endl;

         Charset  set;

         set.set('a', true);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i != 'a'))
                 bomb("one parsed character charset had a mistaken bit set");

         ok("One character Charset test");

    }

    {
         cout << "One character Charset test" << endl;

         Charset  set;

         std::string a("a");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i != 'a'))
                 bomb("one parsed  character charset had a mistaken bit set");

         ok("One parsed character Charset test");

    }

    {
         cout << "One parsed escape character Charset test" << endl;

         Charset  set;

         std::string a("\\t");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i != '\t'))
                 bomb("one parsed escape character charset had a mistaken bit set");

         ok("One parsed escape character Charset test");

    }


    {
         cout << "parsed Minus character Charset test" << endl;

         Charset  set;

         std::string a("-");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i != '-'))
                 bomb("parsed minus character charset had a mistaken bit set");

         ok("Parsed minus character Charset test");

    }

    {
         cout << "Parsed literal string Charset test" << endl;

         Charset  set;

         std::string a("abcd");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i < 'a' || i > 'd' ))
                 bomb("parsed literal charset had a mistaken bit set");

         for(int j='a'; j <= 'd'; ++j)
             if(!set.test(j)) 
                 bomb("parsed literal charset had a mising bit set");


         ok("parsed literal string Charset test");

    }

    {
         cout << "Parsed literal string Charset with escape test" << endl;

         Charset  set;

         std::string a("ab\\bcd");

         set.parse(a);

         if(!set.test(0x08))
         {
                 bomb("parsed literal charset with escape had a missing escaped char");
         }

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i < 'a' || i > 'd' ) && (i != '\b') )
                 bomb("parsed literal charset with escape had a mistaken bit set");

         for(int j='a'; j <= 'd'; ++j)
             if(!set.test(j)) 
                 bomb("parsed literal charset with escape had a mising bit set");


         ok("parsed literal string Charset with escape test");

    }

    {
         cout << "Parsed single range Charset test" << endl;

         Charset  set;

         std::string a("a-d");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i < 'a' || i > 'd' ))
                 bomb("parsed single range had a mistaken bit set");

         for(int j='a'; j <= 'd'; ++j)
             if(!set.test(j)) 
                 bomb("parsed single range had a mising bit set");


         ok("parsed single range Charset test");

    }

    {
         cout << "Parsed escape character range Charset test" << endl;

         Charset  set;

         std::string a("\a-\t");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i) && (i < 0x07 || i > 0x09 ))
                 bomb("parsed escape character range had a mistaken bit set");

         for(int j=0x07; j <= 0x09; ++j)
             if(!set.test(j)) 
                 bomb("parsed escape character range had a mising bit set");


         ok("parsed escape character range Charset test");

    }


    {
         cout << "Parsed minus plus single range Charset test" << endl;

         Charset  set;

         std::string a("-a-d");

         set.parse(a);

         if(!set.test('-'))
         {
            bomb("Parsed minus plus single range is missing the '-' bit!");
         }


         for(int i = 0; i < 256; ++i)
             if(set.test(i) 
                && (
                        (i < 'a')
                    ||  (i > 'd')
                   )
                && ( i != '-' )
               )
                 bomb("minus minus parsed single range had a mistaken bit set");

         for(int j='a'; j <= 'd'; ++j)
             if(!set.test(j)) 
                 bomb("parsed minus plus single range charset had a mising bit set");


         ok("parsed minus plus single range Charset test");

    }

    {
         cout << "Parsed double range Charset test" << endl;

         Charset  set;

         std::string a("a-dqA-Z");

         set.parse(a);

         if(!set.test('q'))
         {
            bomb("Parsed minus plus single range is missing the '-' bit!");
         }

         for(int i = 0; i < 256; ++i)
             if(set.test(i) 
                && (
                        (i < 'a')
                    ||  (i > 'd')
                   )
                && ( i != 'q' )
                && (
                        (i < 'A')
                    ||  (i > 'Z')
                   )
               )
                 bomb("minus minus parsed single range had a mistaken bit set");

         for(int j='a'; j <= 'd'; ++j)
             if(!set.test(j)) 
                 bomb("parsed minus plus single range charset had a mising bit set");

         for(int j='A'; j <= 'Z'; ++j)
             if(!set.test(j)) 
                 bomb("parsed minus plus single range charset had a mising bit set");

         ok("Parsed double range Charset test");

    }

    {
         cout << "Parsed double range (2) Charset test" << endl;

         Charset  set;

         std::string a("qa-dA-Zz");

         set.parse(a);


         for(int i = 0; i < 256; ++i)
             if(set.test(i) 
                && (
                        (i < 'a')
                    ||  (i > 'd')
                   )
                && (
                        (i < 'A')
                    ||  (i > 'Z')
                   )
                && (
                       i != 'z'
                   )
                && (
                       i != 'q'
                   )
               )
                 bomb("minus minus parsed double range (2) had a mistaken bit set");

         for(int j='a'; j <= 'd'; ++j)
             if(!set.test(j)) 
                 bomb("parsed minus plus double range charset (2) had a mising bit set");

         for(int j='A'; j <= 'Z'; ++j)
             if(!set.test(j)) 
                 bomb("parsed minus plus double range charset (2) had a mising bit set");

         if(!set.test('z')) 
                 bomb("parsed minus plus double range charset (2) had a mising bit set");

         if(!set.test('q')) 
                 bomb("parsed minus plus double range charset (2) had a mising bit set");

         ok("Parsed double range Charset (2) test");

    }

    {
         cout << "Parsed square bracket range (2) Charset test" << endl;

         Charset  set;

         std::string a("\\[-\\]");

         set.parse(a);


         for(int i = 0; i < 256; ++i)
             if(set.test(i) 
                && (
                        (i < '[')
                    ||  (i > ']')
                   )
                && (
                       i != '\\'
                   )
               )
                 bomb("minus minus parsed square bracket range (2) had a mistaken bit set");

         for(int j='['; j <= ']'; ++j)
             if(!set.test(j)) 
                 bomb("parsed minus plus square bracket range charset (2) had a mising bit set");

         ok("Parsed square bracket range Charset (2) test");

    }



    {
         cout << "Empty char range Charset test" << endl;

         Charset  set;

         std::string a("");

         set.parse(a);

         for(int i = 0; i < 256; ++i)
             if(set.test(i)) 
                 bomb("empty char range (2) had a mistaken bit set");

         ok("empty char range charset test test");

    }

    {
         cout << "Octal range tests" << endl;

         Charset  set;

         std::string a("\\0-\\027");    // 027 is in the range

         set.parse(a);

         for(int i = 0; i < '\027' + 1; ++i)
             if(!set.test(i)) 
             {
               bomb("octal char range specification failed");
             }

         bool bombed=false;

         for(int i = 0; i < 256; ++i)
             if(set.test(i)) 
             {
                 if(i > '\027' +1 )
                 {
                     bombed = true;
                     cerr << "character " << i << " was not zero!" << endl;
                     cerr << "octal char range specification failed" << endl;
                 }
             }

         if(bombed)  bomb("");


         ok("Octal range tests");
    }

    {
         cout << "Inversion" << endl;

         Charset  set;

         std::string a("^0-9");    // all characters except 0-9

         set.parse(a);

         for(char i = '0'; i <= '9'; ++i)
            if(set.test(i))
            {
               bomb("inverted bit still seemed to be set");
            }

         for(int i=0; i < 256; ++i)
         {
             if( i < '0' || i > '9')
             {
                if(!set.test(i))
                {
                   bomb("inverted unset region was not properly set");
                }
             }
         }


         ok("Inversion");
    }



   ok("ALL CHARSET ");

   return 0;
}
