
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.



//@File
//
///  This file provides a stub implementation of the curses interface.
///  Its purpose is to provide a 'test harness' by which the CursorWindow
///  class can be tested without actually having to press keys on the screen.

#include <iostream>

#include <cxxtls/cursesinterface.h>

using namespace std;
using namespace cxxtls;

namespace cxxtls
{

namespace CursesInterface
{
  bool winch_occurred = false;
  bool resize_requested = false;

  int row;
  int col;

  int att;

  bool kbhit(CursorWindow &)
  {
    return false;
  }

  void openCursesTerminal()
  {
    std::cout << "opening terminal" << std::endl;
  }

  void closeCursesTerminal()
  {
    std::cout << "closing terminal" << std::endl;
  }

  int mapCursesKey(int curses_key)  // map to CursorWindow key representation
  {
    return curses_key;
  }

  void getCursesCursor(int *row, int *col)
  {
    *row = CursesInterface::row;
    *col = CursesInterface::col;
  }

  void getCursesScreenSize(int *row, int *col)
  {
    *row = 25;
    *col = 80;
  }


  void setCursesAttributes( int a )
  {
    CursesInterface::att = a;
  }


  void paintCharString(char const *r, int count, int a, int row, int col)
  {
      std::cout << "paintCharString(\"";

      std::cout.write(r, count);

      std::cout << "\", " << count << ", " << a << ", " << row << ", " << col << ")" << std::endl;

      setCursesAttributes(a);

      moveCursesCursor(row,col);

      while(count--)
          {
              ++CursesInterface::row;
              ++r;
          }
  }

  void paintCharString(long c, int count, int a, int row, int col)
  {
      std::cout << "paintCharString('" << (char)(c) ;
      std::cout << "', " << count << ", " << a << ", " << row << ", " << col << ")" << std::endl;

      setCursesAttributes(a);

      moveCursesCursor(row,col);

      while(count--)
          {
              ++CursesInterface::col;
          }
  }



  void refreshCursesWindow()
  {
  }

  void beepCursesTerminal()
  {
  }


  void moveCursesCursor(int row, int col)
  {
    CursesInterface::row = row;
    CursesInterface::col = col;

    std::cout << "moveCursesCursor(" << row << ", " << col << ")" << std::endl;
  }


  void make_xterm_terminfo_file(FileName const &dir)
  {
  }

    void create_key_sequences_map()
    {
    }


    int alarm_occurred=0;

    int read_key_char()
      // return a 'char' to the user -- or a 0 that indicates the alarm
      // timer fired.
    {
       return ' ';
    }


    int read_mapped_key(CursorWindow &w)
      //
      // read a key from the keyboard and handle function key sequences
      //
    {
      static int eof_counter=0;

      if(std::cin.eof())
      {
         //
         // when encountering an eof situation, alternate between
         // return escape and ' ' -- most of my tests expect ' ' to
         // terminate them and others require escp.
         //
         ++eof_counter;
        
         if(eof_counter & 1)
           return ' ';
         else
           return 0x1b;
      }

      return std::cin.get();

    }

  long line_chars[lcCOUNT];
  mouse_info read_mouse_info() { return mouse_info(0,0,0); } // not yet implemented

};

}
