#include <iostream>
#include <stdlib.h>
#include <cxxtls/patternmatch.h>
#include <cxxtls/charset.h>
#include <memory>
#include <string.h>

using namespace std;
using namespace cxxtls;

void bomb1(char const *msg, char const *file, int line)  { cerr << file << ':' << line << ":     " << msg << endl ; exit(1); }

#define bomb(msg)  bomb1(msg, __FILE__, __LINE__)

#define ok(msg)    cout << "    " << msg << " test passed" << endl;

int main()
{
    cout << "Pattern matcher test begins\n" << endl;


    {
         cout << "Empty matcher test" << endl;

         PatternMatcher  matcher("");

         if(matcher.matches("ab"))
             bomb("empty expression match fails -- non-empty comparand falsely matched");
         else
         if(matcher.matches(""))
             bomb("empty expression match fails -- empty string falsely matches");
         else
             ok("empty expression");


    }

    {
         cout << "Literal string matcher test" << endl;

         PatternMatcher  matcher("abc");

         if(!matcher.matches("abc"))
             bomb("simple literal match fails -- exact match not found");
         else
         if(matcher.matches("123"))
             bomb("simple literal match fails -- wrong data matched anyway");
         else
         if(matcher.matches("axc"))
             bomb("simple literal match fails -- wrong middle char matched anyway");
         else
         if(matcher.matches("abcd"))
             bomb("simple literal match fails -- long comparand falsely matched");
         else
         if(matcher.matches("ab"))
             bomb("simple literal match fails -- short comparand falsely matched");
         else
         if(matcher.matches(""))
             bomb("simple literal match fails -- empty string falsely matches");
         else
             ok("literal text");
             

    }

    {
         cout << "Single ? matcher test" << endl;

         PatternMatcher  matcher("?");

         if(!matcher.matches("a"))
             bomb("single ? match fails -- exact match not found");
         else
         if(matcher.matches("ab"))
             bomb("single ? match fails -- long comparand falsely matched");
         else
         if(matcher.matches(""))
             bomb("single ? match fails -- empty comparand falsely matched");
         else
             ok("single ?");
             

    }

    {
         cout << "Double ? matcher test" << endl;

         PatternMatcher  matcher("??");

         if(!matcher.matches("ab"))
             bomb("double ? match fails -- exact match not found");
         else
         if(matcher.matches("abc"))
             bomb("double ? match fails -- long comparand falsely matched");
         else
         if(matcher.matches(""))
             bomb("double ? match fails -- empty comparand falsely matched");
         else
             ok("double ?");
             

    }

    {
         cout << "Prefix double ? matcher test" << endl;

         PatternMatcher  matcher("a??");

         if(!matcher.matches("abc"))
             bomb("prefix ?? match fails -- exact match not found");
         else
         if(matcher.matches("123"))
             bomb("prefix ?? match fails -- wrong data comparand falsely matched");
         else
         if(matcher.matches("abcd"))
             bomb("prefix ?? match fails -- long comparand falsely matched");
         else
         if(matcher.matches("ab"))
             bomb("prefix ?? match fails -- short comparand falsely matched");
         else
         if(matcher.matches(""))
             bomb("prefix ?? match fails -- empty comparand falsely matched");
         else
             ok("prefix ??");
             

    }

    {
         cout << "Suffix ? matcher test" << endl;

         PatternMatcher  matcher("??c");

         if(!matcher.matches("abc"))
             bomb("postfix ?? match fails -- exact match not found");
         else
         if(matcher.matches("abd"))
             bomb("postfix ?? match fails -- bad last character matched anyway");
         else
         if(matcher.matches("123"))
             bomb("postfix ?? match fails -- wrong data comparand falsely matched");
         else
         if(matcher.matches("abcd"))
             bomb("postfix ?? match fails -- long comparand falsely matched");
         else
         if(matcher.matches("ab"))
             bomb("postfix ?? match fails -- short comparand falsely matched");
         else
         if(matcher.matches(""))
             bomb("postfix ?? match fails -- empty comparand falsely matched");
         else
             ok("postfix ??");
             

    }

    {
         cout << "Only star matcher test" << endl;

         PatternMatcher  matcher("*");

         if(!matcher.matches("a"))
             bomb("star only pattern match fails -- 1 char did not not match");
         else
         if(!matcher.matches("ab"))
             bomb("star only pattern match fails -- 2 chars did not match");
         else
         if(!matcher.matches("123"))
             bomb("star only pattern match fails -- 3 chars did not match");
         else
             ok("star only pattern");

    }

    {
         cout << "Prefix star matcher test" << endl;

         PatternMatcher  matcher("z*");

         if(!matcher.matches("za"))
             bomb("prefix star pattern match fails -- 1 char after prefix did not not match");
         else
         if(!matcher.matches("zab"))
             bomb("prefix star pattern match fails -- 2 chars after prefx did not match");
         else
         if(!matcher.matches("z123"))
             bomb("prefix star pattern match fails -- 3 chars after prefix did not match");
         else
         if(!matcher.matches("z"))
             bomb("prefix star pattern match fails -- prefix did not match");
         else
             ok("prefix star pattern");

    }

    {
         cout << "Postfix star matcher test" << endl;

         PatternMatcher  matcher("*z");

         if(!matcher.matches("az"))
             bomb("postfix star pattern match fails -- 1 char before postfix did not not match");
         else
         if(!matcher.matches("abz"))
             bomb("postfix star pattern match fails -- 2 chars after prefx did not match");
         else
         if(!matcher.matches("123z"))
             bomb("postfix star pattern match fails -- 3 chars before postfix did not match");
         else
         if(!matcher.matches("z"))
             bomb("postfix star pattern match fails -- 0 chars before postfix did not match");
         else
             ok("postfix star pattern");

    }

    
    
    {
         cout << "One char range matcher test" << endl;

         PatternMatcher  matcher("[a-z]");

         if(!matcher.matches("z"))
             bomb("One char range matcher did not match");
         else
         if(!matcher.matches("a"))
             bomb("One char range matcher did not match");
         else
         if(matcher.matches("0"))
             bomb("One char range matcher did not match");
         else
         if(matcher.matches("~"))
             bomb("One char range matcher did not match");
         else
             ok("One char range matcher");

    }

    {
         cout << "Two char range matcher test" << endl;

         PatternMatcher  matcher("[a-z][0-9]");

         if(!matcher.matches("b0"))
             bomb("Two char range matcher did not match");
         else
         if(!matcher.matches("b9"))
             bomb("Two char range matcher did not match");
         else
         if(matcher.matches("b "))
             bomb("Two char range matcher did not match");
         else
         if(matcher.matches("b~"))
             bomb("Two char range matcher did not match");
         else
             ok("Two char range matcher");

    }


    {
         cout << "Prefix two char range matcher test" << endl;

         PatternMatcher  matcher("crap [a-z][0-9]");

         if(!matcher.matches("crap b0"))
             bomb("Prefix two char range matcher did not match");
         else
         if(!matcher.matches("crap b9"))
             bomb("Prefix two char range matcher did not match");
         else
         if(matcher.matches("crap b "))
             bomb("Prefix two char range matcher did not match");
         else
         if(matcher.matches("crap b~"))
             bomb("Prefix two char range matcher did not match");
         else
         if(matcher.matches(""))
             bomb("Prefix two char range matcher did not match");
         else
         if(matcher.matches("crap "))
             bomb("Prefix two char range matcher did not match");
         else
             ok("Prefix two char range matcher");

    }

    {
         cout << "Suffix two char range matcher test" << endl;

         PatternMatcher  matcher("[a-z][0-9].txt");

         if(!matcher.matches("b0.txt"))
             bomb("Suffix two char range matcher did not match");
         else
         if(!matcher.matches("b9.txt"))
             bomb("Suffix two char range matcher did not match");
         else
         if(matcher.matches("b .txt"))
             bomb("Suffix two char range matcher did not match");
         else
         if(matcher.matches("b~.txt"))
             bomb("Suffix two char range matcher did not match");
         else
             ok("Suffix two char range matcher");

    }

    {
         cout << "Star suffix two char range matcher test" << endl;

         PatternMatcher  matcher("*[a-z][0-9].txt");

         if(!matcher.matches("b0.txt"))
             bomb("Star suffix two char range matcher did not match");
         else
         if(!matcher.matches("123b9.txt"))
             bomb("Star suffix two char range matcher did not match");
         else
         if(matcher.matches("b .txt"))
             bomb("Star suffix two char range matcher mistakenly matched");
         else
         if(matcher.matches("abcb~.txt"))
             bomb("Star suffix two char range matcher mistakenly matched");
         else
         if(matcher.matches(""))
             bomb("Star suffix two char range matcher mistakenly matched");
         else
             ok("Star suffix two char range matcher");

    }

    {
       cout << "Multi-star pattern matcher test" << endl;

       PatternMatcher matcher("a*b*c*.txt");

       if(!matcher.matches("abc.txt"))
          bomb("Multi-star pattern match of abc.txt failed");
       else
       if(!matcher.matches("a1b1c1.txt"))
          bomb("Multi-star pattern match of a1b1c1.txt failed");
       else
       if(!matcher.matches("a12b12c12.txt"))
          bomb("Multi-star pattern match of a1b1c1.txt failed");
       else
       if(matcher.matches(""))
          bomb("Multi-star pattern match of empty string mistakenly matched");
       else
       if(matcher.matches("abd.txt"))
          bomb("Multi-star pattern match of empty string mistakenly matched");
       else
       if(matcher.matches("a1b1c1.c"))
          bomb("Multi-star pattern match of a1b1c1.c string mistakenly matched");
       else
           ok("Multi-star pattern match");


    }

    {
       cout << "Multi-star pattern ending in wildcard matcher test" << endl;

       PatternMatcher matcher("a*b*c*");

       if(!matcher.matches("abc"))
          bomb("Multi-star pattern ending in wildcard match of abc failed");
       else
       if(!matcher.matches("a1b1c1"))
          bomb("Multi-star pattern ending in wildcard match of a1b1c1 failed");
       else
       if(!matcher.matches("a12b12c12"))
          bomb("Multi-star pattern ending in wildcard match of a1b1c1 failed");
       else
       if(matcher.matches(""))
          bomb("Multi-star pattern ending in wildcard match of empty string mistakenly matched");
       else
       if(matcher.matches("abd"))
          bomb("Multi-star pattern ending in wildcard match of empty string mistakenly matched");
       else
       if(!matcher.matches("a1b1c1"))
          bomb("Multi-star pattern ending in wildcard match of a1b1c1 string mistakenly matched");
       else
        if(!matcher.matches("abc.txt"))
          bomb("Multi-star pattern ending in wildcard match of abc.txt failed to match");
       else
           ok("Multi-star pattern ending in wildcard match");


    }

    {
       cout << "Multi-star with extra chars pattern ending in wildcard matcher test" << endl;

       PatternMatcher matcher("ax*bx*cx*");

       if(!matcher.matches("axbxcx"))
          bomb("Multi-star with extra chars pattern ending in wildcard match of axbxcx failed");
       else
       if(!matcher.matches("ax1bx1cx1"))
          bomb("Multi-star with extra chars pattern ending in wildcard match of ax1bx1cx1 failed");
       else
       if(!matcher.matches("ax12bx12cx12"))
          bomb("Multi-star with extra chars pattern ending in wildcard match of ax1bx1cx1 failed");
       else
       if(matcher.matches(""))
          bomb("Multi-star with extra chars pattern ending in wildcard match of empty string mistakenly matched");
       else
       if(matcher.matches("axbxdx"))
          bomb("Multi-star with extra chars pattern ending in wildcard match of empty axbxdx mistakenly matched");
       else
        if(!matcher.matches("axbxcx.txt"))
          bomb("Multi-star with extra chars pattern ending in wildcard match of axbxcx.txt failed to match");
       else
        if(matcher.matches("abc.txt"))
          bomb("Multi-star with extra chars pattern ending in wildcard match of abc.txt failed to match");
       else
           ok("Multi-star with extra chars pattern ending in wildcard match");

    }

    {
       cout << "PatternMatcher(*)::shortestLeadingMatch() test" << endl;


       PatternMatcher  matcher("*");

       static char const *comparand[]=
       {
            "",
            "1",
            "1234.z",
            "*",
            "ab\b[1`9]'"
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i];

           size_t l = strlen(p);

           size_t rv = matcher.shortestLeadingMatch(p);

           if(rv != 0)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(*," << p << ") failed to be " << l << endl;
              exit(1);
           }


       }

       ok("PatternMatcher(*)::shortestLeadingMatch()");

    }

    struct PatternAndExpectedLength
    {
       char const *p_;
       size_t expected_;

       PatternAndExpectedLength(char const *p, size_t l)
       :  p_(p)
       ,  expected_(l)
       {
       }

    };



    {
       cout << "PatternMatcher(x)::shortestLeadingMatch() test" << endl;


       PatternMatcher  matcher("x");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "x",                 1),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                1),
            PatternAndExpectedLength( "1x234.z",           0),
            PatternAndExpectedLength( "x1234.z",           1),
            PatternAndExpectedLength( "*",                 0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(x," << p << ") failed to be " << l << endl;
              exit(1);
           }


       }

       ok("PatternMatcher(*)::shortestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(*x)::shortestLeadingMatch() test" << endl;

       PatternMatcher  matcher("*x");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "x",                 1),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                1),
            PatternAndExpectedLength( "1x234.z",           2),
            PatternAndExpectedLength( "x1234.z",           1),
            PatternAndExpectedLength( "*",                 0),
            PatternAndExpectedLength( "fred.x*",           6),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(*x," << p << ") failed to be " << l << endl;
              exit(1);
           }
       }

       ok("PatternMatcher(*x)::shortestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(*xa)::shortestLeadingMatch() test" << endl;

       PatternMatcher  matcher("*xa");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "x",                 0),
            PatternAndExpectedLength( "xa",                2),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                0),
            PatternAndExpectedLength( "xa1",               2),
            PatternAndExpectedLength( "1x234.z",           0),
            PatternAndExpectedLength( "1xa34.z",           3),
            PatternAndExpectedLength( "x1234.z",           0),
            PatternAndExpectedLength( "12x14.z",           0),
            PatternAndExpectedLength( "xa1234.z",          2),
            PatternAndExpectedLength( "3xa124.z",          3),
            PatternAndExpectedLength( "*",                 0),
            PatternAndExpectedLength( "fred.xa*",          7),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(*xa," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }

       ok("PatternMatcher(*xa)::shortestLeadingMatch()");

    }


    {
       cout << "PatternMatcher(x*.)::shortestLeadingMatch() test" << endl;

       PatternMatcher  matcher("x*.");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "x",                 0),
            PatternAndExpectedLength( "x.y",               2),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                0),
            PatternAndExpectedLength( "1x234.z",           0),
            PatternAndExpectedLength( "x1234.z",           6),
            PatternAndExpectedLength( "*",                 0),
            PatternAndExpectedLength( "fred.x*",           0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(x*.," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }

       ok("PatternMatcher(x*.)::shortestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(a?x*.)::shortestLeadingMatch() test" << endl;

       PatternMatcher  matcher("a?x*.");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ax",                0),
            PatternAndExpectedLength( "a1x.y",             4),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                0),
            PatternAndExpectedLength( "a1x234.z",          7),
            PatternAndExpectedLength( "x1234.z",           0),
            PatternAndExpectedLength( "*",                 0),
            PatternAndExpectedLength( "fred.x*",           0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(a?x*.," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }

       ok("PatternMatcher(a?x*.)::shortestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(*a?x*.)::shortestLeadingMatch() test" << endl;

       PatternMatcher  matcher("*a?x*.");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ax",                0),
            PatternAndExpectedLength( "a1x.y",             4),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                0),
            PatternAndExpectedLength( "a1x234.z",          7),
            PatternAndExpectedLength( "x1234.z",           0),
            PatternAndExpectedLength( "*",                 0),
            PatternAndExpectedLength( "fred.x*",           0),

            PatternAndExpectedLength( "2ax",               0),
            PatternAndExpectedLength( "23a1x.y",           6),
            PatternAndExpectedLength( "",                  0),
            PatternAndExpectedLength( "x1",                0),
            PatternAndExpectedLength( "a1x234.z",          7),
            PatternAndExpectedLength( "x1234.z",           0),
            PatternAndExpectedLength( "*a1x.y",            5),
            PatternAndExpectedLength( "freda.x*.y",        9),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch(ab," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }

       ok("PatternMatcher(*a?x*.)::shortestLeadingMatch()");

    }

    {
       cout << "PatternMatcher([a-z4]*[a-z4])::shortestLeadingMatch() test" << endl;

       PatternMatcher  matcher("[a-z4]*[a-z4]");


       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "abc1234@#$",        2),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  shortestLeadingMatch([a-z4]*[a-z4]," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }

       ok("PatternMatcher([a-z4]*[a-z4])::shortestLeadingMatch()");

    }


    {
       cout << "PatternMatcher(ab)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("ab");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           2),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(" << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(ab)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(b)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("b");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           1),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              1),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(b," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(b)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(?)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("?");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           1),
            PatternAndExpectedLength( "123ab",           1),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              1),
            PatternAndExpectedLength( "ab",              1),
            PatternAndExpectedLength( "a",               1),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               1),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(?," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(?)::shortestTrailingMatch()");

    }


    {
       cout << "PatternMatcher(?" "?)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("?""?");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           2),
            PatternAndExpectedLength( "123ab",           2),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(?""?," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(?""?)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher([a-c])::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("[a-c]");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           1),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              1),
            PatternAndExpectedLength( "ab",              1),
            PatternAndExpectedLength( "a",               1),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch([a-c]," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([a-c])::shortestTrailingMatch()");

    }



    {
       cout << "PatternMatcher([a-c][a-d])::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("[a-c][a-d]");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           2),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "bd",              2),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch([a-c][a-d]," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([a-c][a-d])::shortestTrailingMatch()");

    }


    {
       cout << "PatternMatcher(3[a-c][a-d])::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("3[a-c][a-d]");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           3),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             3),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(3[a-c][a-d]," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(3[a-c][a-d])::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(?3[a-c][a-d])::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("?3[a-c][a-d]");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           4),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "z3bd",            4),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(?3[a-c][a-d]," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(?3[a-c][a-d])::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(*)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           0),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "z3bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(*)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(b*)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("b*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           4),
            PatternAndExpectedLength( "123ab",           1),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              1),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             2),
            PatternAndExpectedLength( "z3bd",            2),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(b*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(b*)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(?b*)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("?b*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           2),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             3),
            PatternAndExpectedLength( "z3bd",            3),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch(?b*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(?b*)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher([0-9]?b*)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("[0-9]?b*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           3),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            0),
            PatternAndExpectedLength( "z3bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch([0-9]?b*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([0-9]?b*)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher([0-9]?b*)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("[0-9]?b*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           3),
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            0),
            PatternAndExpectedLength( "z3bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch([0-9]?b*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([0-9]?b*)::shortestTrailingMatch()");

    }

    {
       cout << "PatternMatcher([0-9]*?b*)::shortestTrailingMatch() test" << endl;

       PatternMatcher matcher("[0-9]*?b*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           3),       // matches 3 not 5 because this is a shortest leading match test!
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            0),
            PatternAndExpectedLength( "z3bd",            0),
            PatternAndExpectedLength( "93bd",            4),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.shortestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", shortestTrailingMatch([0-9]*?b*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([0-9]*?b*)::shortestTrailingMatch()");

    }


    {
       cout << "PatternMatcher()::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           0),
            PatternAndExpectedLength( "123ab",           0), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              0),
            PatternAndExpectedLength( "ab",              0),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            0),
            PatternAndExpectedLength( "a1bd",            0),
            PatternAndExpectedLength( "z3bd",            0),
            PatternAndExpectedLength( "93bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch(," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher()::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(*)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           5), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               1),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               1),
            PatternAndExpectedLength( "3bd",             3),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            4),
            PatternAndExpectedLength( "z3bd",            4),
            PatternAndExpectedLength( "93bd",            4),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch(*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(*)::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(?*)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("?*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           5), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               1),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               1),
            PatternAndExpectedLength( "3bd",             3),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            4),
            PatternAndExpectedLength( "z3bd",            4),
            PatternAndExpectedLength( "93bd",            4),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch(?*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(?*)::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(??*)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("??*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           5), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             3),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            4),
            PatternAndExpectedLength( "z3bd",            4),
            PatternAndExpectedLength( "93bd",            4),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch(??*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(??*)::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher([a-z]*)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("[a-z]*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           0), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               1),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               1),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            0),
            PatternAndExpectedLength( "a1bd",            4),
            PatternAndExpectedLength( "z3bd",            4),
            PatternAndExpectedLength( "93bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch([a-z]*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([a-z]*)::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher([a-z][a-z]*)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("[a-z][a-z]*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           0), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            0),
            PatternAndExpectedLength( "a1bd",            0),
            PatternAndExpectedLength( "z3bd",            0),
            PatternAndExpectedLength( "zbbd",            4),
            PatternAndExpectedLength( "93bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch([a-z][a-z]*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([a-z][a-z]*)::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher([a-z]*[a-z]*)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("[a-z]*[a-z]*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           5),
            PatternAndExpectedLength( "123ab",           0), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               0),
            PatternAndExpectedLength( "b",               0),
            PatternAndExpectedLength( "x",               0),
            PatternAndExpectedLength( "3bd",             0),
            PatternAndExpectedLength( "3abd",            0),
            PatternAndExpectedLength( "a1bd",            4),
            PatternAndExpectedLength( "z3bd",            4),
            PatternAndExpectedLength( "zbbd",            4),
            PatternAndExpectedLength( "93bd",            0),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch([a-z]*[a-z]*," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher([a-z]*[a-z]*)::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(*[a-z])::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("*[a-z]");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "ab123",           2),
            PatternAndExpectedLength( "123ab",           5), 
            PatternAndExpectedLength( "",                0),
            PatternAndExpectedLength( "ba",              2),
            PatternAndExpectedLength( "ab",              2),
            PatternAndExpectedLength( "a",               1),
            PatternAndExpectedLength( "b",               1),
            PatternAndExpectedLength( "x",               1),
            PatternAndExpectedLength( "3bd",             3),
            PatternAndExpectedLength( "3abd",            4),
            PatternAndExpectedLength( "a1bd",            4),
            PatternAndExpectedLength( "z3bd",            4),
            PatternAndExpectedLength( "zbbd",            4),
            PatternAndExpectedLength( "93bd",            4),
            PatternAndExpectedLength( "93b3d9",          5),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch(*[a-z]," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(*[a-z])::longestLeadingMatch()");

    }

    {
       cout << "PatternMatcher(*/)::longestLeadingMatch() test" << endl;

       PatternMatcher matcher("*/");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "/ab123.txt",       1),
            PatternAndExpectedLength( "/a/ab123.txt",     3),
            PatternAndExpectedLength( "fred/a/ab123.txt", 7),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestLeadingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestLeadingMatch(*/," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(*/)::longestLeadingMatch()");

    }


    {
       cout << "PatternMatcher(/*)::longestTrailingMatch() test" << endl;

       PatternMatcher matcher("/*");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "/ab123.txt",       10),
            PatternAndExpectedLength( "/a/ab123.txt",     12),
            PatternAndExpectedLength( "fred/a/ab123.txt", 12),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestTrailingMatch(*/," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(*/)::longestTrailingMatch()");

    }

    {
       cout << "PatternMatcher(*.txt)::longestTrailingMatch() test" << endl;

       PatternMatcher matcher("*.txt");

       static PatternAndExpectedLength comparand[]=
       {
            PatternAndExpectedLength( "/ab123.txt",       10),
            PatternAndExpectedLength( "/a/ab123.txt",     12),
            PatternAndExpectedLength( "/a.txt/ab123.txt", 16),
            PatternAndExpectedLength( "/a/ab123-txt",     0),
            PatternAndExpectedLength( "fred/a/ab123.txt", 16),
       };

       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].p_;

           size_t rv = matcher.longestTrailingMatch(p);

           size_t l = comparand[i].expected_;

           if(rv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", longestTrailingMatch(*.txt," << p << ") was " << rv << " not " << l << endl;
              exit(1);
           }
       }
          
       
       ok("PatternMatcher(*.txt)::longestTrailingMatch()");

    }

    {
        cout << "Octal character range in pattern test" << endl;

        PatternMatcher matcher("[\\07-\\11]");

        if(!matcher.matches("\007"))
        {
           bomb("failed to match 007");
        }
        else
        if(!matcher.matches("\010"))
        {
           bomb("failed to match 010");
        }
        else
        if(!matcher.matches("\011"))
        {
           bomb("failed to match 011");
        }

        cout << "    Octal character range in pattern test passed!" << endl;
    }

    {
        cout << "Single Octal in pattern test" << endl;

        PatternMatcher matcher("a\\07b");

        if(!matcher.matches("a\07b"))
        {
           bomb("failed to match 007");
        }

        cout << "    Single Octal character in pattern test passed!" << endl;
    }

    {
        cout << "RemoveLeading test" << endl;

        struct TestData
        {
            char const *pattern, *value, *longest, *shortest;

            TestData(char const *p, char const *v, char const *l, char const *s)
            :  pattern(p)
            ,  value(v)
            ,  longest(l)
            ,  shortest(s)
            {
            }
        };

        static TestData comparand[]=
        {
                                                 //longest  //shortest
             TestData("fred", "frederick",       "erick",   "erick"),
             TestData("f*e",  "frederick",       "rick",    "derick"),
             TestData("f*e",  "thomas",          "thomas",  "thomas"),
             TestData("*/",    "/fred/bil.c",    "bil.c",   "fred/bil.c"),
        };


       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].pattern;
           char const *c = comparand[i].value;
           char const *l = comparand[i].longest;
           char const *s = comparand[i].shortest;

           PatternMatcher matcher(p);

           std::string lrv = matcher.removeLeadingMatch(c, true);

           if(lrv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", removeLongestLeadingMatch(" << p << "," << c << ") was " << lrv << " not " << l << endl;
              exit(1);
           }

           std::string srv = matcher.removeLeadingMatch(c);

           if(srv != s)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", removeShortestLeadingMatch(" << p << "," << c << ") was " << srv << " not " << s << endl;
              exit(1);
           }

       }



        cout << "    RemoveLeading test PASSED" << endl;
    }

    {
        cout << "RemoveTrailing test" << endl;

        struct TestData
        {
            char const *pattern, *value, *longest, *shortest;

            TestData(char const *p, char const *v, char const *l, char const *s)
            :  pattern(p)
            ,  value(v)
            ,  longest(l)
            ,  shortest(s)
            {
            }
        };

        static TestData comparand[]=
        {
                                                //longest   //shortest
             TestData("fred", "thomas",         "thomas",   "thomas"),
             TestData("fred", "manfred",        "man",      "man"),
             TestData("d*y",  "freddy",         "fre",      "fred"),

        };


       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].pattern;
           char const *c = comparand[i].value;
           char const *l = comparand[i].longest;
           char const *s = comparand[i].shortest;

           PatternMatcher matcher(p);

           std::string lrv = matcher.removeTrailingMatch(c, true);

           if(lrv != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", removeLongestTrailingMatch(" << p << "," << c << ") was " << lrv << " not " << l << endl;
              exit(1);
           }

           std::string srv = matcher.removeTrailingMatch(c);

           if(srv != s)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", removeShortestTrailingMatch(" << p << "," << c << ") was " << srv << " not " << s << endl;
              exit(1);
           }

       }



        cout << "    RemoveTrailing test PASSED" << endl;
    }

////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

    {
        cout << "find without offset test" << endl;

        struct TestData
        {
            char const *pattern, *value;

            size_t expected, length;

            TestData(char const *p, char const *v, size_t e, size_t l)
            :  pattern(p)
            ,  value(v)
            ,  expected(e)
            ,  length(l)
            {
            }
        };

        static TestData comparand[]=
        {
                       
             TestData("x", "", -1, 0),
             TestData("x", "x", 0, 1),
             TestData("x", "ax", 1, 1),
             TestData("x", "xa", 0, 1),


             TestData("xy", "xya", 0, 2),
             TestData("xy", "axya", 1, 2),
             TestData("xy", "aaxy", 2, 2),

             TestData("x?y", "aaxzy", 2, 3),
             TestData("x?y", "xzy",   0, 3),
             TestData("x?y", "xy",   -1, 0),

        };


       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].pattern;
           char const *v = comparand[i].value;
           size_t      e = comparand[i].expected;
           size_t      l = comparand[i].length;

           PatternMatcher matcher(p);

           PatternMatcher::OffsetAndLength rv = matcher.find(v);

           if(rv.first != e)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", findWithoutOffset(" << p << "," << v << ") had offset= " << rv.first << " not " << e << endl;
              exit(1);
           }

           if(rv.second != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", findWithoutOffset(" << p << "," << v << ") had length= " << rv.second << " not " << l << endl;
              exit(1);
           }


       }



        cout << "    find without offset test PASSED" << endl;
    }

    {
        cout << "find with offset test" << endl;

        struct TestData
        {
            char const *pattern, *value;

            size_t expected, length;

            TestData(char const *p, char const *v, size_t e, size_t l)
            :  pattern(p)
            ,  value(v)
            ,  expected(e)
            ,  length(l)
            {
            }
        };

        static TestData comparand[]=
        {
                       
             TestData("x", "****", -1, 0),
             TestData("x", "****x", 4, 1),
             TestData("x", "****ax", 5, 1),
             TestData("x", "****xa", 4, 1),


             TestData("xy", "****xya", 4, 2),
             TestData("xy", "****axya", 5, 2),
             TestData("xy", "****aaxy", 6, 2),

             TestData("x?y", "****aaxzy", 6, 3),
             TestData("x?y", "****xzy",   4, 3),
             TestData("x?y", "****xy",   -1, 0),

        };


       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].pattern;
           char const *v = comparand[i].value;
           size_t      e = comparand[i].expected;
           size_t      l = comparand[i].length;

           PatternMatcher matcher(p);

           PatternMatcher::OffsetAndLength rv = matcher.find(v,4);

           if(rv.first != e)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", findWithOffset(" << p << "," << v << ") had offset= " << rv.first << " not " << e << endl;
              exit(1);
           }

           if(rv.second != l)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i << ", findWithOffset(" << p << "," << v << ") had length= " << rv.second << " not " << l << endl;
              exit(1);
           }


       }



        cout << "    find with offset test PASSED" << endl;
    }

    {
        cout << "Replacment test" << endl;

        struct TestData
        {
            char const *pattern, *value, *replacement, *expected;

            bool all;


            TestData(char const *p, char const *v, char const *r, bool a, char const *e)
            :  pattern(p)
            ,  value(v)
            ,  replacement(r)
            ,  expected(e)
            ,  all(a)
            {
            }
        };

        static TestData comparand[]=
        {
                       
             TestData("x",   "", "X", 1, "" ),
             TestData("x",   "x", "X", 1, "X" ),
             TestData("x",   "xx", "X", 1, "XX" ),
             TestData("x",   "xx", "X", 0, "Xx" ),
             TestData("xx",  "xxx", "X", 1, "Xx" ),

             TestData("x*x", "xxx", "", 1, "" ),
             TestData("x*x", "(xxx)", "", 1, "()" ),



        };


       for(size_t i = 0; i < sizeof(comparand) / sizeof(comparand[0]); ++i)
       {
           char const *p = comparand[i].pattern;
           char const *v = comparand[i].value;
           char const *r = comparand[i].replacement;
           char const *e = comparand[i].expected;

           bool all      = comparand[i].all;

           PatternMatcher matcher(p);

           std::string rv = matcher.replaceMatches(v, r, all);

           if(rv != e)
           {
              cerr << __FILE__ << ':' << __LINE__ << ":  at pattern " << i 
                   << ", Replacment(" << p << "," << v << "," << r << "," << (all? 1 : 0) << ") found " << rv << ", not " << e << endl;
              exit(1);
           }

       }
        cout << "    Replacment test PASSED" << endl;
    }

    {
        cout << "Backslash test" << endl;

        PatternMatcher matcher("\\");

        if(!matcher.matches("\\"))
        {
           bomb("a single backslash character crashes the pattern matcher parser");
        }


        cout << "    Backslash test  PASSED" << endl;
    }



    cout << "All tests have passed!" << endl;
}
