
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.



#include <file>

#include <iostream>

int main(int argc, char **argv)
{
  if(argc < 2)
    return 0;

  FileName f(argv[1]);

  FileTime time;

  FileName ap(f);

  f.file_time(&time);

  ap.convert_to_absolute_path();

  FileStatus fs;

  f.file_stat(&fs);

  std::cout << "name=       " << f                          << std::endl
	    << "owner=      " << fs.id2string(fs.owner_id)  << std::endl
	    << "valid=      " << f.is_syntactically_valid() << std::endl
	    << "readable=   " << f.is_readable()            << std::endl
	    << "writeable=  " << f.is_writeable()           << std::endl
	    << "createable= " << f.is_createable()          << std::endl
	    << "exists=     " << f.exists()                 << std::endl
	    << "executable= " << f.is_executable()          << std::endl
	    << "directory=  " << f.is_dir()                 << std::endl
	    << "mode=       " << std::oct << f.file_mode()  << std::dec << std::endl
	    << "size=       " << f.file_size()              << std::endl
	    << "time=       " << time                       << std::endl
	    << "dirname=    " << f.dirname()                << std::endl
	    << "basename=   " << f.basename()               << std::endl
	    << "absolute=   " << ap                         << std::endl
	    << "absdirname= " << ap.dirname()               << std::endl
	    << "absbasename=" << ap.basename()              << std::endl
	    ;


  { // test the os_mode to mode conversions

    FileMode perm1("drwxrwxrwx");

    FileMode perm2( perm1.os_mode() );

    if(perm1.mode != perm2.mode )
      std::cout << "Oops, FileMode::os_mode() didn't work right" << std::endl;


    FileName tmpdir = FileName::mktemp();

    FileMode dirperms("dr-x------");


    if(tmpdir.mkdir(dirperms))
      std::cout << "Error, couldn't create directory " << tmpdir << std::endl;
    else
    {
       FileStatus fs;

       if(tmpdir.file_stat(&fs))
	 std::cout << "Error, couldn't read file status of " << tmpdir << std::endl;
       {
	 if(fs.mode.mode != dirperms.mode)
	   std::cout << "Error, newly created directory has bad permissions" << std::endl;
       }

       if(tmpdir.rmdir())
       {
	 std::cout << "Error, couldn't remove directory " << tmpdir << std::endl;
       }

    }


  }


  return 0;

}

