
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/ftp.h>
#include <stdlib.h>

using namespace std;
using namespace cxxtls;

int main()
{
   {
     FileStatus stat;

     cout << "Validing parsing of ftp directory lines:  1" << endl;

     char const* dirtext=
	 "drwxrwxrwx   1 owner    group   9 Feb 10 2003           incoming\n";

     FileName filename;

     if(!FTP::parse_ftp_dir_text(dirtext, &stat, &filename) )
     {
       cout << "Doh!  can't understand example ftp directory text!" << endl;
     }
     else
       cout << stat << " " << filename << endl;

     if( (1900 + stat.time.year) != 2003)
     {
	cout << "Doh!  FTP::parse_ftp_dir_text got the wrong year" << endl;
	exit(1);
     }

     if(stat.time.month != 1) // feb
     {
	cout << "Doh!  FTP::parse_ftp_dir_text got the wrong month" << endl;
	exit(1);
     }

     if(stat.time.day != 10)
     {
	cout << "Doh!  FTP::parse_ftp_dir_text got the wrong day" << endl;
	exit(1);
     }

     if(!stat.mode.is_dir())
     {
	cout << "Doh!  FTP::parse_ftp_dir_text got the permissions wrong" << endl;
	exit(1);
     }

   }


   {
     FileStatus stat;

     cout << "Validing parsing of ftp directory lines:  2" << endl;

     char const* dirtext=
       "-rwxrwxrwx   1 owner    group        79982792 Feb 19  9:07 scp_engine.6.0.new\n";

     FileName filename;

     if(!FTP::parse_ftp_dir_text(dirtext, &stat, &filename) )
     {
       cout << "Doh!  can't understand example ftp directory text!" << endl;
     }
     else
     {
       if(stat.time.year > 103 && stat.time.year < 200)
       {
	 stat.time.year = 103;
       }
       cout << stat << " " << filename << endl;
     }

     if(stat.mode.is_dir())
     {
	cout << "Doh!  FTP::parse_ftp_dir_text got the permissions wrong" << endl;
	exit(1);
     }

     if(stat.time.hour != 9)
     {
	cout << "Doh!  FTP::parse_ftp_dir_text got the hour wrong" << endl;
	exit(1);
     }


   }

   {
     FileStatus stat;

     cout << "Validing parsing of ftp directory lines:  3" << endl;

     char const* dirtext=
	 "drwSr-xr--   1 owner    group   9 Dec 14 2013           outgoing\n";

     FileName filename;

     if(!FTP::parse_ftp_dir_text(dirtext, &stat, &filename) )
     {
       cout << "Doh!  can't understand example ftp directory text!" << endl;
     }
     else
       cout << stat << " " << filename << endl;
   }


   {
     cout << "Listing a subdirectory of ftp.gnu.org" << endl;

     std::list<FTP::FileInfo> stat;

     if( 0 < FTP::stat_matching("ftp.gnu.org",
			"anonymous",
			"anonymous@junk.com",
			"welcome*",
			&stat
		       ) 
       )
     {
       cout << "Cool, it passed!" << endl;
     }
     else
     {
       cout << "Oops, failed:  " << FTP::error() << endl;
     }
   }

   {
     std::list<FTP::FileInfo> stat;

     cout << "Making sure stuff that doesn't exist, doesn't" << endl;

     if(!FTP::stat_matching("ftp.gnu.org",
			"anonymous",
			"anonymous@junk.com",
			"crapola92",
			&stat
		       ) &&
	stat.size() != 0
       )
     {
       cout << "Huh, this should have failed" << endl;
     }

   }

}
