
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.



#include <cxxtls/options.h>
#include <memory.h>
#include <stdlib.h>
#include <portable_io.h>
#include <iostream>

//@File
//
///  This file defines a test program that validates the functionality
///  of classes defined in include/options
//
//


using namespace std;
using namespace cxxtls;

int main()
{
    static char const *environ[]=
    {
      "env1=v1",
      "env2=v2",
      0
    };

    static char const *arg[]=
    {
      "program name",
      "-x", // 3 invokes of -x
      "-x",
      "-x",
      "-y", // 2 invokes of -y
      "-y",
      "-I", // 3 invokes of -I
      "filename",
      "-I../include",
      "-I", // 3 invokes of -I
      "another",
      "parm1",
      "parm2",
      "parm3",
      0
    };


    static char const *other_arg[]=
    {
      "program name",
      "-one", // 2 invokes of -I
      "first",
      "-one", // 2 invokes of -I
      "second",
      "parm1",
      "parm2",
      "parm3",
      0
    };


    ProgramOptions::Descriptors d;

    d.add_option("-x");
    d.add_option("-y");
    d.add_option("-one", ProgramOptions::one_value);
    d.add_option("-I", ProgramOptions::many_values);

    ProgramOptions op(d, -1 + sizeof(arg)/sizeof(char*), arg, environ);
    ProgramOptions other(d, -1 + sizeof(other_arg)/sizeof(char*), other_arg, environ,false);

    ProgramOptions sop("-x,-y,-one:-I;", -1 + sizeof(arg)/sizeof(char*), arg, environ);



    ProgramOptions::Environment const &env = op.get_environment();

    {
      cout << "Environment is: " << endl;


      ProgramOptions::Environment::const_iterator f = env.begin();
      ProgramOptions::Environment::const_iterator l = env.end();

      while(f != l)
      {
         string const& name = f->first;
         string const& value= f->second;
        
         cout << "    " << name << "=" << value << endl;
        
         ++f;
        
      }
    }

    {
      cout << "Non-Option arguments are: " << endl;

      size_t i;

      for(i=0; i < op.argv.size(); ++i)
        cout << "    " << op.argv[i] << endl;
        
        
      if(sop.argv.size() != op.argv.size())
        cout << "Huh?  sop.argv.size() is wrong" << endl;
        
      for(i=0; i < op.argv.size(); ++i)
      {
        if(sop.argv[i] != op.argv[i])
        {
          cout << "Huh?  sop.argv[" << i << "] is wrong" << endl;
          break;
        }
      }

    }

    {
      cout << "Optional arguments are: " << endl;

      ProgramOptions::Options::const_iterator f = op.get_options().begin();
      ProgramOptions::Options::const_iterator l = op.get_options().end();

      ProgramOptions::Options::const_iterator m = sop.get_options().begin();
      ProgramOptions::Options::const_iterator n = sop.get_options().end();

      while(f != l)
      {
         cout << "    " << f->second->name() << "=" << (string)(*f->second) << endl;
        
         if(m->second->name() != f->second->name())
         {
           cout << "Huh?  sop.options don't match op.options()" << endl;
           break;
         }
        
         ++f;
         ++m;

      }

      if(m != n)
      {
        cout << "Huh?  sop.options don't match op.options()" << endl;
      }


    }

    if(op.option("-x"))
      cout << "The -x option was correctly set" << endl;
    else
      cout << "Huh? where's the -x option" << endl;

    if(op.option("-y"))
      cout << "Huh?  -y option was doubly set and should not appear" << endl;


    if(!sop.option("-x"))
      cout << "Huh? where's the -x option in sop" << endl;

    if(sop.option("-y"))
      cout << "Huh?  -y option was doubly set in sop and should not appear" << endl;

    if(!other.error())
      cout << "Huh?  other_arg should have an error in it" << endl;
    else
      cout << "The error test, other_arg, gives error " << other.error_text() << endl;

    if(!op.getenv("env1"))
      cout << "Huh? environment variable env1 is not set correctly" << endl;
    else
      cout << "env1's value is " << op.getenv("env1").value() << endl;

    if(!op.getenv("env2"))
      cout << "Huh? environment variable env2 is not set correctly" << endl;
    else
      cout << "env2's value is " << op.getenv("env2").value() << endl;


    if(op.getenv("env3"))
      cout << "Huh?  Why is there an env3 variable set?" << endl;

    string exp = op.expand("abc-$3-$env2-");

    if(exp != "abc-parm3-v2-")
      cout << "Expansion wrong, got " << exp << ", instead of 'abc-parm3-v2-'"
           << endl;
    else
      cout << "Expansion of 'abc-$3-$env2-' gave " << exp << endl;

    exp = op.expand("$0$1$2$3$env2$env1");

    if(exp != "program nameparm1parm2parm3v2v1")
      cout << "Expansion wrong, got " << exp << endl;
    else
      cout << "Expansion of '$0$1$2$3$env2$env1' gave " << exp << endl;



   exit(0);

}



