
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.



#include <string.h>
#include <portable_io.h>
#include <stdlib.h>
#include <cxxtls/simple_regex.h>
#include <iostream>

using namespace	std;
using namespace cxxtls;

int main(int argc, char **argv)
{
    int rc;
    int i;

    if(argc < 3)
	{
	    fprintf(stderr, "error, neet pattern and test string\n");
	    exit(1);
	}


    SimpleRegex r(argv[1]);

    if( !r.ok() )
    {
      printf("regex compilation error:  '%s'\n", r.error());
      exit(1);
    }


    rc = r(argv[2], strlen(argv[2]));

    printf("searchRegexp(%s,%s) = %d\n", argv[1], argv[2], rc);

    for(i=0; i < rc; ++i)
    {
	printf("match[%d] begins at %d and is %d bytes long\n",
	       i,
	       r.matches()[i].offset,
	       r.matches()[i].length
	      );
    }

    {
      SimpleRegex r1("bill");

      printf("Now testing substitute\n\n");

      string original = "some bill needs payment";
      string expected =	"some [BILL] needs payment";

      int count	= r1(original);

      if(count < 1)
      {
	fprintf(stderr,	"Dude!	a trivial regex	find failed!\n");
	exit(1);
      }

      string rep1 = r1.substitute(original, "[BILL]");

      string rep = rep1;

      if(rep ==	original)
      {
	fprintf(stderr,	"Dude, like no replacement occurred\n");
	exit(1);
      }

      if(rep !=	expected)
      {
	cerr << "Dude, like incorrect replacement occurred" << endl ;
	cerr << "expected '" << expected << "'" << endl;
	cerr << "but got  '" << rep << "'" << endl;
        exit(1);
      }

      printf("Yeah!, found %s\n", rep.c_str());


    }


    {
      SimpleRegex r1("\\(bill\\) \\(needs\\)");

      printf("Now testing a more complex substitute\n\n");

      string original = "some bill needs payment";
      string expected = "some [bill][needs] payment";

      int count = r1(original);

      if(count < 1)
      {
	fprintf(stderr, "Dude!  a parenthetical regex find failed!\n");
	exit(1);
      }

      string rep1 = r1.substitute(original, "[\\1][\\2]");

      string rep = rep1;

      if(rep == original)
      {
	fprintf(stderr, "Dude, like no replacement occurred\n");
	exit(1);
      }

      if(rep != expected)
      {
	cerr << "Dude, like incorrect replacement occurred" << endl ;
	cerr << "expected '" << expected << "'" << endl;
	cerr << "but got  '" << rep << "'" << endl;
	exit(1);
      }

      printf("Yeah!, found %s\n", rep.c_str());


    }

  return 0;
}

