#include <cxxtls/resourcepool.h>
#include <iostream>
#include <stdlib.h>
using namespace std;

#define ASSUMED(v1,eq,v2) \
    if( !((v1) eq (v2)) ) \
    { \
      cout << __FILE__ << ":" << __LINE__ << ":  Error, expected " << #v1 << " " << #eq << " " << #v2 << endl; \
      cout << "    but found " << v1 << endl; \
      cout << "    And       " << v2 << endl; \
      cout << "TEST FAILED!" << endl; \
      exit(1); \
    }

using namespace cxxtls;

int main()
{

   cout << "RESOURCE POOL EXACT MATCH TEST" << endl;
   {
        ResourcePool<int,int> rp;
        rp.adopt(  new int(1), 1 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(5), 5 );

        ScopedBorrow<int,int>  b0(rp, 0);  // should not find a match
        ASSUMED(b0.get(), ==, 0);

        ScopedBorrow<int,int>  b1(rp, 1);  // should get the single 1
        ASSUMED(b1.get(), !=, 0);

        ScopedBorrow<int,int>  b1b(rp, 1);  // should not get another 1
        ASSUMED(b1b.get(), ==, 0);

        ScopedBorrow<int,int>  b2(rp, 2);  // should not find a match
        ASSUMED(b2.get(), ==, 0);

        ScopedBorrow<int,int>  b3a(rp, 3);  // should get one of the two
        ASSUMED(b3a.get(), !=, 0);

        ScopedBorrow<int,int>  b3b(rp, 3);  // should get the second of the two
        ASSUMED(b3b.get(), !=, 0);

        ScopedBorrow<int,int>  b3c(rp, 3);  // should not find a theird 
        ASSUMED(b3c.get(), ==, 0);

        ScopedBorrow<int,int>  b4(rp, 4);  // should not find a theird 
        ASSUMED(b4.get(), ==, 0);

        ScopedBorrow<int,int>  b5(rp, 5);  // should find the single 5
        ASSUMED(b5.get(), !=, 0);

        ScopedBorrow<int,int>  b6(rp, 6);  // should not find a match
        ASSUMED(b6.get(), ==, 0);


   }

   cout << "RESOURCE POOL <= MATCH TEST" << endl;
   {
        ResourcePool<int,int> rp;
        rp.adopt(  new int(1), 1 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(5), 5 );

        ScopedBorrow<int,int> b0(rp, 0, false, true);  // attemping to borrow 0 or less should fail.
        ASSUMED(b0, ==, 0);       

        ScopedBorrow<int,int> b6(rp, 6, false, true);  // attemping to borrow 6, with dn == true will give you 5
        ASSUMED(b6, !=, 0);  
        ASSUMED(*b6, ==, 5);

        ScopedBorrow<int,int> b6b(rp, 6, false, true);  // attemping to borrow 6, with dn == true will give you 5
        ASSUMED(b6b, !=, 0);  
        ASSUMED(*b6b, ==, 3);

        ScopedBorrow<int,int> b2(rp, 2, false, true);  // attemping to borrow 6, with dn == true will give you 5
        ASSUMED(b2, !=, 0);  
        ASSUMED(*b2, ==, 1);

        ScopedBorrow<int,int> b2b(rp, 2, false, true);  // attemping to borrow 6, with dn == true will give you 5
        ASSUMED(b2b, ==, 0);  

        ScopedBorrow<int,int> b3(rp, 3, false, true);  // attemping to borrow 6, with dn == true will give you 5
        ASSUMED(b3, !=, 0);  
        ASSUMED(*b6b, ==, 3);

   }

   cout << "RESOURCE POOL >= MATCH TEST" << endl;
   {
        ResourcePool<int,int> rp;
        rp.adopt(  new int(1), 1 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(5), 5 );

        ScopedBorrow<int,int> b0(rp, 0, true);
        ASSUMED(b0.get(), !=, 0);
        ASSUMED(*b0,      ==, 1);

        ScopedBorrow<int,int> b1(rp, 1, true);
        ASSUMED(b1.get(), !=, 0);
        ASSUMED(*b1,      ==, 3);

        ScopedBorrow<int,int> b2(rp, 2, true);
        ASSUMED(b2.get(), !=, 0);
        ASSUMED(*b2,      ==, 3);

        ScopedBorrow<int,int> b6(rp, 6, true);
        ASSUMED(b6.get(), ==, 0);

        ScopedBorrow<int,int> b3(rp, 3, true);
        ASSUMED(b3.get(), !=, 0);
        ASSUMED(*b3,      ==, 5);

        ScopedBorrow<int,int> b4(rp, 4, true);
        ASSUMED(b4.get(), ==, 0);


   }

   cout << "RESOURCE BORROW/RELEASE" << endl;
   {
        ResourcePool<int,int> rp;
        rp.adopt(  new int(1), 1 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(3), 3 );
        rp.adopt(  new int(5), 5 );

        {
             {
                 ScopedBorrow<int,int> b3(rp,3);
                 ASSUMED(*b3, ==, 3);
             
                 ScopedBorrow<int,int> b3a(rp,3);
                 ASSUMED(*b3a, ==, 3);
             
                 ScopedBorrow<int,int> b3b(rp,3);
                 ASSUMED(b3b, ==, 0);
             
             }
             
             {
                 ScopedBorrow<int,int> b3(rp,3);
                 ASSUMED(*b3, ==, 3);
             
                 ScopedBorrow<int,int> b3a(rp,3);
                 ASSUMED(*b3a, ==, 3);
             
                 ScopedBorrow<int,int> b3b(rp,3);
                 ASSUMED(b3b, ==, 0);
             
             }
             
             ScopedBorrow<int,int> b5(rp, 5);
             ASSUMED(b5, !=, 0);
             ASSUMED(*b5, ==, 5);
             
             {
                 ScopedBorrow<int,int> b3(rp,3);
                 ASSUMED(*b3, ==, 3);
             
                 ScopedBorrow<int,int> b3a(rp,3);
                 ASSUMED(*b3a, ==, 3);
             
                 ScopedBorrow<int,int> b3b(rp,3);
                 ASSUMED(b3b, ==, 0);
             
                 ScopedBorrow<int,int> b5(rp, 5);
                 ASSUMED(b5, ==, 0);
             
             
             
             }
        }


        ScopedBorrow<int,int>  b0(rp, 0);  // should not find a match
        ASSUMED(b0.get(), ==, 0);

        ScopedBorrow<int,int>  b1(rp, 1);  // should get the single 1
        ASSUMED(b1.get(), !=, 0);

        ScopedBorrow<int,int>  b1b(rp, 1);  // should not get another 1
        ASSUMED(b1b.get(), ==, 0);

        ScopedBorrow<int,int>  b2(rp, 2);  // should not find a match
        ASSUMED(b2.get(), ==, 0);

        ScopedBorrow<int,int>  b3a(rp, 3);  // should get one of the two
        ASSUMED(b3a.get(), !=, 0);

        ScopedBorrow<int,int>  b3b(rp, 3);  // should get the second of the two
        ASSUMED(b3b.get(), !=, 0);

        ScopedBorrow<int,int>  b3c(rp, 3);  // should not find a theird 
        ASSUMED(b3c.get(), ==, 0);

        ScopedBorrow<int,int>  b4(rp, 4);  // should not find a theird 
        ASSUMED(b4.get(), ==, 0);

        ScopedBorrow<int,int>  b5(rp, 5);  // should find the single 5
        ASSUMED(b5.get(), !=, 0);

        ScopedBorrow<int,int>  b6(rp, 6);  // should not find a match
        ASSUMED(b6.get(), ==, 0);


   }


   
   cout << "ResourcePool test passed" << endl;
   exit(0);
}
