#include <cxxtls\scriptthread.h>
#include <stdlib.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

using namespace std;
using namespace cxxtls;
using namespace ScriptThread;

    struct JobCat   // a test job.  reads from stdin and writes the same to stdout until end of file
    :  public Job
    {

        JobCat(ThreadContext *adoptee)
        :  Job(adoptee)
        {
        }

        virtual void *run();  // copy adoptee->stdin to adoptee->stdout until end of file then quit.

    private:
        ~JobCat() {}

    };

    struct JobByteCount   // a test job.  reads from stdin until end of file.  Counts the number of bytes
                          // read and writes the number to stdout as a string with a newline at the end.
    :  public Job
    {

        JobByteCount(ThreadContext *adoptee)
        :  Job(adoptee)
        {
        }

        virtual void *run();  // read from stdin till end of file.  write the byte count to stdout with a newline

     private:
        ~JobByteCount() {}

    };

    struct JobPrintBytes   // a test job.  prints a specified number of bytes to stdout followed by a newline.
    :  public Job
    {
        int count_;


        JobPrintBytes(ThreadContext *adoptee, int count)
        :  Job(adoptee)
        ,  count_(count)
        {
        }

        virtual void *run();  // read from stdin till end of file.  write the byte count to stdout with a newline

     private:
        ~JobPrintBytes() {}

    };



void runTest(ManagedFileHandles *handles, char const *redirFile, char const *runMessage)
{
    ManagedFileHandles::IO_ERROR err;

    StringExpr::ExecutionVariables emptyVariableSet;


    string helloWorld(":  hello standard out:   being written to '");

    if( (err = handles->writeBytes(1, string(runMessage) + helloWorld + string(redirFile) + string("'\n"))) )
    {
        cout << "Error:  an error occurred when writing to mainStdOut;  " << ManagedFileHandles::errorString(err) << endl;
    }

    if( (err = handles->writeBytes(2, "hello standard error\n")) )
    {
        cout << "Error:  an error occurred when writing to mainStdErr;  " << ManagedFileHandles::errorString(err) << endl;
    }

    cout << "\nEnter some text" << endl;

    char buffer[40];
    size_t count=0;

    if( (err = handles->readBytes(0, 8, buffer, &count)) )
    {
        cout << "Error:  an error occurred when reading from mainStdin;  " << ManagedFileHandles::errorString(err) << endl;
    }

    cout << "Read " << count << " bytes:  '";

    for(size_t i = 0; i < count; ++i)  cout << buffer[i];

    cout << "'" << endl;

    if( !handles->reopenAs(1, redirFile) )
    {
        cout << "Error:  cannot reopen stdout out as " << redirFile << endl;
    }
    else
    {
      if( (err = handles->writeBytes(1, std::string("hello ") + std::string(redirFile) + "\n")) )
      {
          cout << "Error:  an error occurred when writing to redirected mainStdOut;  " << ManagedFileHandles::errorString(err) << endl;
      }
      else
      {

         int processError = -99;


         processError = 
#ifdef _MSC_VER
         handles->callProcess("cmd /c dir /b Makefile", emptyVariableSet);  
#else
         handles->callProcess("ls Makefile", emptyVariableSet);
#endif

         if( (err = handles->writeBytes(1, "you should see 'Makefile' above this line\n") ) )
         {
             cout << "Error:  an error occurred when writing to redirected mainStdOut;  " << ManagedFileHandles::errorString(err) << endl;
         }


         if(processError < 0)
         {
             if( (err = handles->writeBytes(1, "ERROR:  callProcess() failed!  \n") ) )
             {
                 cout << "Error:  an error occurred when writing to redirected mainStdOut;  " << ManagedFileHandles::errorString(err) << endl;
             }
         }


      }
    }

}


#ifdef _MSC_VER

char* iGetLastErrorText(DWORD nErrorCode)
{
    char* msg;
    // Ask Windows to prepare a standard message for a GetLastError() code:

    int lcMessage = 
    FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, nErrorCode, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)&msg, 0, NULL);
    // Return the message
    if (!lcMessage)
        return("Unknown error");
    else
        return(msg);
}

#endif


int main()
{
    cout << "Beginning the test using the default file handle set" << endl;
    cout << "=======================================================================" << endl;
    runTest(ManagedFileHandles::mainStdHandles, "st_test.redir1.txt", "First Invocation");
    cout << "Test Complete" << endl;
    cout << "=======================================================================" << endl;
    cout << "Repeating Test using cloned file handles" << endl;
    cout << "=======================================================================" << endl;

    // at this point, the default stdandard output handle has been redirected to "st_test.redir1.txt", 
    // so the results of the second run will be slightly different than the first.  The message "hello standard out"
    // will go to file st_test.redir1.txt instead of stdout.

    cout << "Starting the clone() test" << endl;

    ManagedFileHandles* clone = ManagedFileHandles::mainStdHandles->clone();

    // NOTE:  some text written by the test will go to st_test.redir1.txt before the test
    // redirects stdout to st_test.redir2.txt.  Don't have a cow over it man!

    runTest(clone, "st_test.redir2.txt", "CloneTest");

    cout << "Confirming that that the pre-cloned file handles still work" << endl;

    ManagedFileHandles::mainStdHandles->writeBytes(1,"After clone test:  This should end up in st_test.redir1.txt\n");



    cout << "Test Complete" << endl;
    cout << "=======================================================================" << endl;

    cout << "Verifying that reopenAs() works on stdin" << endl;
    cout << "=======================================================================" << endl;

    if(!clone->reopenAs(0, "st_test.in.other"))
    {
       cout << "Error:  could not open st_test.in.other" << endl;
    }
    else
    {
        char buffer[40];

        size_t bytes_read = 0;

        if(clone->readBytes(0, sizeof(buffer), buffer, &bytes_read))
        {
            cout << "Error:  could not read from file st_test.in.other after redirecting standard input from there" << endl;
        }
        else
        {
            buffer[bytes_read] = 0;

            cout << "Read '" << buffer << "' from st_test.in.other" << endl;
        }
    }

    cout << "Cloning the std file handles 20,000 times" << endl;
    
    for(int i = 0; i < 20000; ++i)
    {
        ManagedFileHandles*  p = ManagedFileHandles::mainStdHandles->clone();

        if(!p)
        {
           cerr << "Error:  could not open all the managed file handles" << endl;
           cout << "Error:  could not open all the managed file handles" << endl;
           exit(1);
        }


        ManagedFileHandles::destroy(p);
    }

    {
        cout << "Test piping" << endl;
        
        ManagedFileHandles *   child  = ManagedFileHandles::mainStdHandles->clone();
        if(!child)
        {
           cerr << "Error:  could not clone mainStdHandles " << endl;
           exit(1);
        }
        
        ManagedFileHandles *   source = child->pipe();

        if(!source)
        {
           cerr << "Error:  could not establish a pipe" << endl;
           exit(1);
        }

        source->writeBytes(1,"x\r\n");

        char buffer[40];

        size_t readCount=0;

        int rc = child->readBytes(0,40, buffer,&readCount );

        if(rc != 0)
        {
           cerr << "Could not read the 3 bytes written to the pipe" << endl;
           exit(1);
        }

        cout << "Read " << readCount << " expected 3 byte" << endl;

        if(buffer[0] == 'x')
        {
           cout << "Yea! pipe data read correctly" << endl;
        }
        else
        {
           cout << "Doh!  Pipe data is bogus" << endl;
        }


        ManagedFileHandles::destroy(child);
        ManagedFileHandles::destroy(source);
    }

    {
       cout << "Test thread pipes" << endl;

        ManagedFileHandles *   child  = ManagedFileHandles::mainStdHandles->clone();
        if(!child)
        {
           cerr << "Error:  could not clone mainStdHandles " << endl;
           exit(1);
        }

        if(!child->reopenAs(1,"threadChild.txt") )
        {
           cerr << "Error:  unable to reopen stdout in cloned ManagedFileHandles to text file"
                << endl;
           exit(1);
        }

        ManagedFileHandles *   source = child->pipe();

        if(!source)
        {
           cerr << "Error:  could not establish a pipe" << endl;
           exit(1);
        }

        cout << "Writing one line to child pipe" << endl;

        source->writeBytes(1,"x\r\n");

        cout << "Creating a thread to read it" << endl;

        StringExpr::ExecutionVariables  environmentForJob;;


        ThreadContext *childContext = ThreadContext::create("cat",
                                                            StringExpr::ExecutionContext::ArgsVector(),
                                                            &environmentForJob, 
                                                            child
                                                           );
                                      
                                                                                               

        JobCat *jc = new JobCat(childContext);  // child will read "x" newline then quit, and
                                                // destroy childContext

        cout <<  "Waiting for the job to complete" << endl;

        // ThreadId  tid = createThread(&JobCat::start,jc);

        ThreadId tid = jc->start();

        ManagedFileHandles::destroy(source); // send end of file to jc

        joinThread(tid);  // wait till thread terminates


    }

    {
       cout << "Test thread pipe volume operations" << endl;

        ManagedFileHandles *   child  = ManagedFileHandles::mainStdHandles->clone();
        if(!child)
        {
           cerr << "Error:  could not clone mainStdHandles " << endl;
           exit(1);
        }

        if(!child->reopenAs(1,"threadChild2.txt") )
        {
           cerr << "Error:  unable to reopen stdout in cloned ManagedFileHandles to text file"
                << endl;
           exit(1);
        }

        ManagedFileHandles *   source = child->pipe();

        if(!source)
        {
           cerr << "Error:  could not establish a pipe" << endl;
           exit(1);
        }

        cout << "Creating a thread to read from the pipe" << endl;

        StringExpr::ExecutionVariables environmentForJob;

        ThreadContext *childContext = ThreadContext::create("lc",
                                                            StringExpr::ExecutionContext::ArgsVector(),
                                                            &environmentForJob, 
                                                            child
                                                           );
                                      
                                                                                               

        JobByteCount *jc = new JobByteCount(childContext);  // child count the bytes in its input stream.
                                                      // destroy childContext

        // ThreadId  tid = createThread(&JobByteCount::start,jc);

        ThreadId tid = jc->start();


        cout << "Writing 50,000 lines to child pipe" << endl;

        for(int i = 0; i < 50000; ++i)
            source->writeBytes(1,"x\r\n");

        cout <<  "Waiting for the job to complete" << endl;


        ManagedFileHandles::destroy(source); // send end of file to jc

    }

    {
       cout << "Test thread to thread piping volume operations" << endl;

        ManagedFileHandles *   child  = ManagedFileHandles::mainStdHandles->clone();
        if(!child)
        {
           cerr << "Error:  could not clone mainStdHandles " << endl;
           exit(1);
        }

        if(!child->reopenAs(1,"threadChild3.txt") )
        {
           cerr << "Error:  unable to reopen stdout in cloned ManagedFileHandles to text file"
                << endl;
           exit(1);
        }

        ManagedFileHandles *   source = child->pipe();

        if(!source)
        {
           cerr << "Error:  could not establish a pipe" << endl;
           exit(1);
        }

        cout << "Creating a thread to read from the pipe" << endl;

        StringExpr::ExecutionVariables lcEnvironment;


        ThreadContext *readerContext = ThreadContext::create("lc",
                                                            StringExpr::ExecutionContext::ArgsVector(),
                                                            &lcEnvironment,
                                                            child
                                                           );
                                      
                                                                                               

        JobByteCount *jc = new JobByteCount(readerContext);  // child count the bytes in its input stream.
                                                      // destroy readerContext

        // ThreadId  tid = createThread(&JobByteCount::start,jc);

        ThreadId tid = jc->start();


        cout << "Creating a thread to write to the pipe" << endl;

        StringExpr::ExecutionVariables nlEnvironment;


        ThreadContext *writerContext = ThreadContext::create("nl",
                                                            StringExpr::ExecutionContext::ArgsVector(),
                                                            &nlEnvironment, 
                                                            source
                                                           );
                                      

        JobPrintBytes *jw = new JobPrintBytes(writerContext, 200 * 1000 );  // print 200 thousan bytes followed by \r\n

        ThreadId wtid = jw->start();

        cout <<  "Waiting for the job to complete" << endl;

        joinThread(wtid);  // wait for the writer to finish writing its data
        joinThread(tid);   // wait for the reader to print out the contents -- the size in bytes of the data written
                                                                            

    }

    
    cout << "Now, the Makefile will include include the extra test file outputs for confirmation." << endl;



    // put nothing here!
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   void *
   JobCat::
   run()
   {
       char buffer[8192];

       if(context_->fileHandles_->numHandles() < 2 )
         return 0;  // huh?

       for(;;)
       {
           size_t bytesRead=0;

           context_->fileHandles_->readBytes(0, sizeof(buffer), buffer, &bytesRead);


           if(bytesRead == 0)
              break;
           else
           {
              if(!context_->fileHandles_->writeBytes(1, buffer, bytesRead))
                break;
           }
       }


       return 0;
   }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   void *
   JobByteCount::
   run()
   {
       char buffer[8192];

       if(context_->fileHandles_->numHandles() < 2 )
         return 0;  // huh?

       long long total = 0;

       for(;;)
       {
           size_t bytesRead=0;

           context_->fileHandles_->readBytes(0, sizeof(buffer), buffer, &bytesRead);


           if(bytesRead == 0)
              break;
           else
           {
              total += bytesRead;
           }
       }


       snprintf(buffer, sizeof(buffer), "Total bytes read %lld\n", total);


       context_->fileHandles_->writeBytes(1, buffer, strlen(buffer));


       return 0;
   }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

   void *
   JobPrintBytes::
   run()
   {
       char buffer[81];

       if(context_->fileHandles_->numHandles() < 2 )
         return 0;  // huh?

       int value=1;

       while(count_--)
       {
           
           context_->fileHandles_->writeBytes(1, "z", 1);

       }

#ifdef _MSC_VER
       context_->fileHandles_->writeBytes(1, "\r\n", 1);
#else
       context_->fileHandles_->writeBytes(1, "\n", 1);
#endif


       return 0;
   }

