
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.



#include <iostream>
#include <cxxtls/strtool.h>
#include <cxxtls/file.h>
#include <cxxtls/foreach.h>
#include <cxxtls/csv.h>


//@File
///  Test 'strtool' header definitions and a few in the 'file'
///  header.
//

using namespace std;
using namespace cxxtls;

int main()
{

  {
      string fragment("BC");
      string within("abc");

      size_t offset = StrTool::find(fragment, within, true);

      if(offset != 1)
      {
         cout << "Darn it!  StrTool::find(BC,abc,true) failed!";

      }

      offset = StrTool::find(fragment, within, false);

      if(offset < within.size())
      {
         cout << "Darn it!  StrTool::find(BC,abc,false) failed!";
      }


      offset = StrTool::find("bc", within, true);

      if(offset != 1)
      {
         cout << "Darn it!  StrTool::find(\"bc\",abc,true) failed!";

      }


  }

  {
    StrTool::stringlist_t l;

    int rv = StrTool::parse_words("a b c d", &l);

    if(rv != 4)
    {
       cout << "Whoa dude, like the parsing of 4 simple words did not work" << endl;
    }
    else
    {
       StrTool::stringlist_t::const_iterator n = l.begin();
       StrTool::stringlist_t::const_iterator e = l.end();

       while(n != e)
       {
         cout << *n++ << endl;
       }

    }
  }

  {
    StrTool::stringlist_t l;

    int rv = StrTool::parse_words("", &l);

    if(rv != 0 || l.begin() != l.end() )
    {
       cout << "Whoa dude, like the parsing of 0 simple words did not work" << endl;
    }
  }

  {
    StrTool::stringlist_t l;

    int rv = StrTool::parse_words("a1234 b c d   ", &l, 3);

    if(rv != 3)
    {
       cout << "Whoa dude, like parsing 2 words and leaving the rest in a string"
            << " did not work.  The returned count is " << rv << endl;
    }
    else
    {
       StrTool::stringlist_t::const_iterator n = l.begin();
       StrTool::stringlist_t::const_iterator e = l.end();

       while(n != e)
       {
         cout << "'" << *n++ << "'" << endl;
       }

    }
  }

  FileName::mktemp();           // just here so I can use the debugger
  FileName::mktemp("stuff");    // to check proper operation

  {
    string PATH= string("../tests")       +
                 FileName::PATH_separator +
                 string("../bin")
                 ;

    FileName f = FileName::find_executable("test_strtool.exe", PATH);

    cout << "Found test_strtool.exe in '" << f << "'" << endl;


  }

  {  // test tab expansion and compression

     string rv = StrTool::expand_tabs("stuff");

     if(rv != "stuff")
     {
       cout << "expand_tabs is broken in a major way" << endl;
     }

     rv = StrTool::expand_tabs("    xyz");

     if(rv != "    xyz")
     {
       cout << "expand_tabs is broken in a subtle way" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("\tx");

     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs(" \tx");

     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("  \tx");

     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("   \tx");

     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("    \tx");

     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("     \tx");

     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("      \tx");

     if(rv != "        x" )
     {                              
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
                        
     rv= StrTool::expand_tabs("       \tx");
        
     if(rv != "        x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }


     rv= StrTool::expand_tabs("y\tx");
                
     if(rv != "y       x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("y\t\tx");
                
     if(rv != "y               x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv= StrTool::expand_tabs("123456789\tx");
                        
     if(rv != "123456789       x" )
     {
       cout << "expand_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv = StrTool::pack_tabs("1234567890");

     if(rv != "1234567890" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv = StrTool::pack_tabs(" 1234567890");

     if(rv != " 1234567890" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv = StrTool::pack_tabs("12 34567890");

     if(rv != "12 34567890" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }


     rv = StrTool::pack_tabs("12  34567890");

     if(rv != "12  34567890" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
                           //  12345678
     rv = StrTool::pack_tabs("0        x",false);

     if(rv != "0\t x" )
     {
       cout << "pack_tabs is broken, found '" << rv << "', ( " << __FILE__ << ":" << __LINE__ << endl;
     }
                           //  123456789
     rv = StrTool::pack_tabs("0\t  x",false);

     if(rv != "0\t  x" )
     {
       cout << "pack_tabs is broken, found '" << rv << "', ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv = StrTool::pack_tabs("\t\t\tx");

     if(rv != "\t\t\tx" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }

     rv = StrTool::pack_tabs("  \t  \t  \t  x");

     if(rv != "\t\t\t  x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
                            //12345678
     rv = StrTool::pack_tabs("      x");
             //1234567
     if(rv != "      x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
       int count = 0; 

       while(rv[count] == ' ') ++count;

       cout << "There were  " << count << " blanks " << endl;
       cout << "followed by " << (int)rv[count] << " decimal " << endl;
       cout << "which is    " << rv[count] << endl;
       cout << "then        " << rv[count+1] << endl;

     }

                            //12345678
     rv = StrTool::pack_tabs("       x");
             //1234567890
     if(rv != "       x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;


       int count = 0; 

       while(rv[count] == ' ') ++count;

       cout << "There were  " << count << " blanks " << endl;
       cout << "followed by " << (int)rv[count] << " decimal " << endl;
       cout << "which is    " << rv[count] << endl;
       cout << "then        " << rv[count+1] << endl;


     }
     else
       cout << rv << endl;

                            //12345678
     rv = StrTool::pack_tabs("        x");

     if(rv != "\tx" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
     else
       cout << rv << endl;
                            //123456789
     rv = StrTool::pack_tabs("12345678x");

     if(rv != "12345678x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
     else
       cout << rv << endl;


                            //12345678901234
     rv = StrTool::pack_tabs("12345678 x");

     if(rv != "12345678 x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
     else
       cout << rv << endl;


                            //123456789
     rv = StrTool::pack_tabs("1234567 x");

     if(rv != "1234567 x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;


       int count = 0; 

       while(rv[count+7] == ' ') ++count;

       cout << "  There were  " << count << " blanks " << endl;
       cout << "  followed by " << (int)rv[count+7] << " decimal " << endl;
       cout << "  which is    " << rv[count+7] << endl;
       cout << "  then        " << rv[count+1+7] << endl;

     }
     else
       cout << rv << endl;


                            //12345678901234567890
     rv = StrTool::pack_tabs("12345678   x");
             //1234567890123456789
     if(rv != "12345678   x" )
     {
       cout << "pack_tabs is broken" << " ( " << __FILE__ << ":" << __LINE__ << endl;
     }
     else
       cout << rv << endl;


  }  // tabs tests

  {  // snatch_file_info

      
      static string grepOutput = "filename:12:";                  // grep -n 
      static string borland    = "file 13:";                      // borland C output
      static string ibm        = "\"file\", line 2:";             // other compiler outputs
      static string microsoft  = "file(14)";                      // microsoft C output
      static string ansiCPP    = "#line 12 \"file\"";             // ANSI C preprocessor
      static string include    = "<file>";                        // C language #include file names
      static string gccCPP     = "#12 \"file\"";                  // gcc's preprocessor output
      static string ibmErr     = "Error 112: \"reports_pkg.c\", line 48";    // ibm error msg format
      static string hpErr      = "Error (112): \"reports_pkg.c\", line 49";  // hp error msg format
      static string gccErr     = "error: \"reports_pkg.c\", line 48";        // gcc error
      static string gccWarn    = "warning: \"reports_pkg.c\", line 50"; // gcc warning
      static string grepOut2   = "file:";                         // grep without the -n
      static string valgrind1  = "(file.cxx:949)";                // valgrind logs with a debug compile
      static string valgrind2  = "(in file.cxx:)";                // valgrind logs with an optimized compile



      struct tableEntry
      {
          string title_;
          string text_;
          string expectedFile_;
          int    expectedLine_;
      };

      static tableEntry tests[]=
      {
         { "grepOutput", grepOutput, "filename",      12 },
         { "borland",    borland,    "file",          13 },
         { "ibm",        ibm,        "file",          2  },
         { "microsoft",  microsoft,  "file",          14 },
         { "ansiCPP",    ansiCPP,    "file",          12 },
         { "include",    include,    "file",          1  },
         { "gccCPP",     gccCPP,     "file",          12 },
         { "ibmErr",     ibmErr,     "reports_pkg.c", 48 },
         { "hpErr",      hpErr,      "reports_pkg.c", 49 },
         { "gccErr",     gccErr,     "reports_pkg.c", 48 },
         { "gccWarn",    gccWarn,    "reports_pkg.c", 50 },
         { "grepNo_N",   grepOut2,   "file",          1  },

         { "valgrind1",  valgrind1,   "file.cxx",     949},
         { "valgrind2",  valgrind2,   "file.cxx",     1  },
      };

      for(size_t i = 0; i < sizeof(tests)/ sizeof(tests[0]); ++i)
      {

          tableEntry const &test = tests[i];

          string filename;
          int    lineNo=-1;

          bool success = StrTool::snatch_file_info(test.text_, filename, lineNo);

          if(!success)
          {
             cout << "StrTool::snatch_file_info() failed on format " << test.title_ << endl;
             cout << "  parsed values are file='" << filename << "', and line=" << lineNo << endl;
             exit(1);
          }

          if(   (filename != tests[i].expectedFile_)
             || (lineNo  != tests[i].expectedLine_ )
            )
          {
             cout << "StrTool::snatch_file_info() failed on format " << test.title_ << endl;
             cout << "  parsed values are file='" << filename << "', and line=" << lineNo << endl;
             exit(1);
          }
            


      }

      cout << "Snatch file info test passed" << endl;


  }  // snatch_file_info

  {
       string testLine("abc,,def,,,h");

       static char const *expected[]=
       {
          "abc",
          "",
          "def",
          "",
          "",
          "h"
       };

       vector<string> output;

       size_t count = StrTool::splitOnDelimiter(testLine,output,StrTool::Is_Comma());

       if(count != 6)
       {
          cout << "StrTool::splitOnDelimiter():   did not find 6 fields, it found " << count << " instead" << endl;
          exit(1);
       }

       if(output.size() != 6)
       {
          cout << "StrTool::splitOnDelimiter():   output vector did not contain 6 fields, it found " << output.size() << " instead" << endl;
          exit(1);
       }

       for(int i = 0; i < 6; ++i)
       {
          if(expected[i] != output[i])
          {
              cout << "StrTool::splitOnDelimiter():  output[" << i << "] did not contain '" << expected[i] << "', it contained '" << output[i] << "' instead" << endl;
              exit(1);
          }
       }


  }

  {
       string testLine("");

       vector<string> output;

       size_t count = StrTool::splitOnDelimiter(testLine,output,StrTool::Is_Comma());

       if(count != 0)
       {
          cout << "StrTool::splitOnDelimiter():   did not find 0 fields, it found " << count << " instead" << endl;
          exit(1);
       }

       if(output.size() != 0)
       {
          cout << "StrTool::splitOnDelimiter():   output vector did not contain 0 fields, it found " << output.size() << " instead" << endl;
          exit(1);
       }

  }

  {
       string testLine("fred");

       vector<string> output;

       size_t count = StrTool::splitOnDelimiter(testLine,output,StrTool::Is_Comma());

       if(count != 1)
       {
          cout << "StrTool::splitOnDelimiter():   did not find 1 fields, it found " << count << " instead" << endl;
          exit(1);
       }

       if(output.size() != 1)
       {
          cout << "StrTool::splitOnDelimiter():   output vector did not contain 1 fields, it found " << output.size() << " instead" << endl;
          exit(1);
       }

       if(output[0] != "fred")
       {
          cout << "StrTool::splitOnDelimiter():   output vector[0] did not contain 'fred' fields, it had '" << output[0] << "' instead" << endl;
          exit(1);
       }

  }

  {
   std::vector<std::string> vs;

   std::string data("a,13,'stuff,'',crap',b,-\"\"\"\"-,\"\"\"\",''''");

   if(!csv::split(data, vs))
   {
      std::cerr << "Failed the parse test:  (" << data << ")" << std::endl;

      for(size_t i = 0; i < vs.size(); ++i)
          std::cout << "[" << i << "] = " << vs[i] << std::endl;

   }
   else
   {
         static char const *expected[]=
         {
             "a",
             "13",
             "stuff,',crap",
             "b",
             "-\"-",
             "\"",
             "'"
         };

       size_t expectedCount = sizeof(expected) / sizeof(expected[0]);

      if(vs.size() != expectedCount)
         std::cerr << "Failed the parse test: (" << data << ")" << " found " << vs.size() << " of " << expectedCount << " fields " << std::endl;
      else
      {
         for(size_t i = 0; i < vs.size(); ++i)
         {
            if(vs[i] != expected[i])
            {
               std::cout << "Error:  at expected location " << i << " found (" << vs[i] << ") instead of " << expected[i] << std::endl;
            }
         }

      }

   }
     
  }

  {
      list<string>  container;

      container.push_back("a");
      container.push_back(",");
      container.push_back("a,b");
      container.push_back("'c'");
      container.push_back("\"");

      string formatted;

      csv::format(container, formatted);

      string expected("a,',','a,b','''c''','\"'");

      if(formatted != expected)
      {
         std::cerr << "Error:  csv format failure:\n"
                   <<"   expeced:  " << expected << "\n"
                   <<"     found:  " << formatted 
                   ;
      }

  }



  exit(0);
}
