#include <cxxtls/timeseries.h>
#include <cxxtls/stlstream.h>
#include <cxxtls/foreach.h>
#include <set>
#include <cfloat>

using namespace std;
using namespace cxxtls;

int main()
{

   cout << "======================================================" << endl;
   cout << "Single Square wave correlation example: " << endl;

   {
       vector<int>   vi;
       
       vi.push_back(0);
       vi.push_back(1);
       vi.push_back(0);
       vi.push_back(1);
       vi.push_back(0);
       vi.push_back(1);
       
       cout << "time series= " << vi << endl;
       cout << "mean of vi = " << arithmeticMean(vi) << endl;

       for(int i = 0; i < int(vi.size()-1); ++i)
       {
           cout << "autoCorrelationFunction(" << i << ") = " << autoCorrelationFunction(vi,i) << endl;
       }

       {
           TimeSeriesAnalysis tsa(vi);
           
           cout << "TSA(VI) time deltas" << endl;
           
           CXXTLS_FOREACH(size_t cur, tsa.acfPeriodWidths_)
           {
              cout << "    " << cur << endl;
           }
       }


   }
   cout << "======================================================" << endl;
   cout << "double wide Square wave correlation example: " << endl;

   {
       vector<int>   vi;
       
       vi.push_back(0);
       vi.push_back(0);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(0);
       vi.push_back(0);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(0);
       vi.push_back(0);
       vi.push_back(1);
       vi.push_back(1);


       cout << "time series= " << vi << endl;
       cout << "mean of vi = " << arithmeticMean(vi) << endl;


       vector<int> maxima;

       vector<long double> acf;

       for(int i = 0; i < int(vi.size()-1); ++i)
       {
           long double value = autoCorrelationFunction(vi,i);

           acf.push_back(value);

           cout << "autoCorrelationFunction(" << i << ") = " << autoCorrelationFunction(vi,i) << endl;

       }

       cout << "acf mean = " << arithmeticMean(acf) << ", stdDev = " << standardDeviation(acf) << endl;
       cout << "acf min  = " << minimum(acf) << ", max = " << maximum(acf) << endl;

       bool goingDown = true;
       long double prev    = acf[0];
       size_t prevI   = 0;

       typedef map<long double, size_t> ACF_LEVEL_TYPE;
       typedef ACF_LEVEL_TYPE::value_type ACF_LEVEL_ENTRY;


       ACF_LEVEL_TYPE    acfLevels;
       vector<long double>    acfLevelValues;

       acfLevels.insert(ACF_LEVEL_ENTRY(1.0,0));

       for(size_t i = 1; i < acf.size(); ++i)
       {
          long double current = acf[i];

          if(goingDown)
          {
              if(current <= prev)
              {
                 prev = current;
                 continue;
              }
              else
                 goingDown = false; // ie we are now going up
          }
          else
          {
             // going up

             if(current >= prev)
             {
                prev = current;
                continue;
             }
             else
             {
                // prev is a local maximum
                goingDown = true;  // now we are falling
                prevI = i -1;
                cout << "Found a maximum at i = " << prevI << ", value = " << prev << endl;

                acfLevels.insert( ACF_LEVEL_ENTRY(prev,prevI) );
                acfLevelValues.push_back(prev);
             }

          }
       }

       cout << "levels of acf maxima: " << endl;


       CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, acfLevels)
       {

          cout << "    " << cur << endl;
          
       }

       long double acfMean, acfLevelStdDev;

       cout << "ACF Level Mean = " << (acfMean = arithmeticMean(acfLevelValues)) << endl;
       cout << "ACF StandardDev= " << (acfLevelStdDev = standardDeviation(acfLevelValues)) << endl;

       multiset<size_t> timeIndices;

       long double minLevel = 1 - acfLevelStdDev;
       long double levelSlop = 1e-7;

       minLevel -= levelSlop;

       {
           
           
           
           CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, acfLevels)
           {
              if(cur.first  >= minLevel)
              {
                  timeIndices.insert(cur.second);
              }
              
           }

           cout << "Main sequence maxima:" << endl;
           CXXTLS_FOREACH(size_t cur, timeIndices)
           {
              cout << "    " << cur << endl;
           }
           
           vector<size_t> timeDeltas;
           
           cout << "computing deltas in the time index series: " << endl;

           size_t largestDelta = 0;
           
           {
               multiset<size_t>::iterator first = timeIndices.begin(), last = timeIndices.end();
           
               while(first != last)
               {
                  size_t cur = *first++;
           
                  if(first != last)
                  {
                     long double delta = *first - cur;
           
                     timeDeltas.push_back(delta); 
           
                     if(delta > largestDelta)
                        largestDelta = delta;

                     cout << "    [" << *first << ", " << cur << "] = " << delta << endl;
                  }
           
           
               }
           }

           size_t meanTimeDelta = ceil(arithmeticMean(timeDeltas));
           size_t meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           
           cout << "Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;


           cout << "Removing large items" << endl;


           while(largestDelta > meanTimeStdDev)
           {
               // remove items out of range

               vector<size_t> tmp;

               largestDelta = 0;

               CXXTLS_FOREACH(size_t cur, timeDeltas)
               {
                   if(cur < meanTimeStdDev)
                   {
                      if(cur > largestDelta)
                         largestDelta = cur;

                      tmp.push_back(cur);
                   }

               }

               if(tmp.empty())
                  break;

               cout << "timeDeltas = " << timeDeltas << endl;
               cout << "tmp        = " << tmp        << endl;

               timeDeltas.swap(tmp);

               cout << "after swap timeDeltas = " << timeDeltas << endl;

               // recompute statistics

               meanTimeDelta = ceil(arithmeticMean(timeDeltas));
               meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           }


           cout << "Filtered Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Filtered Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;

           largestDelta = maximum(timeDeltas);

           cout << "Removing Time deltas less than " << largestDelta - meanTimeStdDev << endl;

           {
              vector<size_t> tmp;

              CXXTLS_FOREACH(size_t cur, timeDeltas)
              {
                 if(cur >= largestDelta - meanTimeStdDev)
                 {
                    tmp.push_back(cur);
                 }
              }

              if(tmp.size())
                 timeDeltas.swap(tmp);

           }

           cout << "Filtered Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Filtered Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;

           cout << "Final time deltas" << endl;

           CXXTLS_FOREACH(size_t cur, timeDeltas)
           {
              cout << "    " << cur << endl;
           }




       }  // main sequence

       {
           TimeSeriesAnalysis tsa(vi);
           

           cout << "levels of acf maxima: " << endl;
    
    
           CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, tsa.acfPeriods_)
           {
    
              cout << "    " << cur << endl;
              
           }

           cout << "ACF Level Mean = " << tsa.acfLevelMean_   << endl;
           cout << "ACF StandardDev= " << tsa.acfLevelStdDev_ << endl;


           cout << "Main sequence maxima:" << endl;

           CXXTLS_FOREACH( size_t cur, tsa.acfMainSequenceTimeIndices_)
           {
              cout << "    " << cur << endl;
           }




           cout << "TSA(VI) time period widths" << endl;


           CXXTLS_FOREACH(size_t cur, tsa.acfPeriodWidths_)
           {
              cout << "    " << cur << endl;
           }
       }


   }


   cout << "======================================================" << endl;
   cout << "Repeating ramp example" << endl;

   {
       vector<int>   vi;
       
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(30);
       

       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(30);


       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(1);
       vi.push_back(30);

       cout << "time series= " << vi << endl;
       cout << "mean of vi = " << arithmeticMean(vi) << endl;

       for(int i = 0; i < int(vi.size()-1); ++i)
       {
           cout << "autoCorrelationFunction(" << i << ") = " << autoCorrelationFunction(vi,i) << endl;
       }

       {
           TimeSeriesAnalysis tsa(vi);
           
           cout << "TSA(VI) time deltas" << endl;
           
           CXXTLS_FOREACH(size_t cur, tsa.acfPeriodWidths_)
           {
              cout << "    " << cur << endl;
           }
       }


   }
   cout << "======================================================" << endl;
   cout << "Pittsburg hourly temp measurements (oct 1 - 4, 2013)" << endl;
   {
       vector<int>   vi;

       vi.push_back(65  );
       vi.push_back(64  );
       vi.push_back(64  );
       vi.push_back(63  );
       vi.push_back(62  );
       vi.push_back(62  );
       vi.push_back(62  );
       vi.push_back(62  );
       vi.push_back(60  );
       vi.push_back(59  );
       vi.push_back(57  );
       vi.push_back(58  );
       vi.push_back(58  );
       vi.push_back(60  );
       vi.push_back(65  );
       vi.push_back(68  );
       vi.push_back(70  );
       vi.push_back(72  );
       vi.push_back(73  );
       vi.push_back(74  );
       vi.push_back(75  );
       vi.push_back(76  );
       vi.push_back(75  );
       vi.push_back(72  );
       vi.push_back(69  );
       vi.push_back(68  );
       vi.push_back(66  );
       vi.push_back(64  );
       vi.push_back(64  );
       vi.push_back(63  );
       vi.push_back(63  );
       vi.push_back(59  );
       vi.push_back(59  );
       vi.push_back(58  );
       vi.push_back(60  );
       vi.push_back(60  );
       vi.push_back(61  );
       vi.push_back(63  );
       vi.push_back(67  );
       vi.push_back(71  );
       vi.push_back(74  );
       vi.push_back(76  );
       vi.push_back(76  );
       vi.push_back(77  );
       vi.push_back(80  );
       vi.push_back(80  );
       vi.push_back(78  );
       vi.push_back(75  );
       vi.push_back(73  );
       vi.push_back(69  );
       vi.push_back(68  );
       vi.push_back(66  );
       vi.push_back(64  );
       vi.push_back(63  );
       vi.push_back(61  );
       vi.push_back(61  );
       vi.push_back(60  );
       vi.push_back(60  );
       vi.push_back(60  );
       vi.push_back(60  );
       vi.push_back(62  );
       vi.push_back(63  );
       vi.push_back(65  );
       vi.push_back(70  );
       vi.push_back(73  );
       vi.push_back(74  );
       vi.push_back(76  );
       vi.push_back(76  );
       vi.push_back(78  );
       vi.push_back(78  );
       vi.push_back(77  );
       vi.push_back(75  );
       vi.push_back(68  );
       vi.push_back(69  );
       vi.push_back(67  );
       vi.push_back(68  );
       vi.push_back(66  );
       vi.push_back(65  );
       vi.push_back(64  );
       vi.push_back(64  );
       vi.push_back(64  );
       vi.push_back(63  );
       vi.push_back(65  );
       vi.push_back(64  );
       vi.push_back(65  );
       vi.push_back(66  );
       vi.push_back(69  );
       vi.push_back(69  );
       vi.push_back(77  );
       vi.push_back(79  );
       vi.push_back(80  );
       vi.push_back(80  );
       vi.push_back(78  );
       vi.push_back(72  );
       vi.push_back(73  );
       vi.push_back(73  );
       vi.push_back(71  );
       vi.push_back(70  );
       vi.push_back(70  );
       vi.push_back(70  );


       cout << "time series= " << vi << endl;
       cout << "mean of vi = " << arithmeticMean(vi) << endl;


       vector<int> maxima;

       vector<long double> acf;

       for(int i = 0; i < int(vi.size()-1); ++i)
       {
           long double value = autoCorrelationFunction(vi,i);

           acf.push_back(value);

           cout << "autoCorrelationFunction(" << i << ") = " << autoCorrelationFunction(vi,i) << endl;

       }

       cout << "acf mean = " << arithmeticMean(acf) << ", stdDev = " << standardDeviation(acf) << endl;
       cout << "acf min  = " << minimum(acf) << ", max = " << maximum(acf) << endl;

       bool goingDown = true;
       long double prev    = acf[0];
       size_t prevI   = 0;

       typedef map<long double, size_t> ACF_LEVEL_TYPE;
       typedef ACF_LEVEL_TYPE::value_type ACF_LEVEL_ENTRY;


       ACF_LEVEL_TYPE    acfLevels;
       vector<long double>    acfLevelValues;

       acfLevels.insert(ACF_LEVEL_ENTRY(1.0,0));

       for(size_t i = 1; i < acf.size(); ++i)
       {
          long double current = acf[i];

          if(goingDown)
          {
              if(current <= prev)
              {
                 prev = current;
                 continue;
              }
              else
                 goingDown = false; // ie we are now going up
          }
          else
          {
             // going up

             if(current >= prev)
             {
                prev = current;
                continue;
             }
             else
             {
                // prev is a local maximum
                goingDown = true;  // now we are falling
                prevI = i -1;
                cout << "Found a maximum at i = " << prevI << ", value = " << prev << endl;

                acfLevels.insert( ACF_LEVEL_ENTRY(prev,prevI) );
                acfLevelValues.push_back(prev);
             }

          }
       }

       cout << "levels of acf maxima: " << endl;


       CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, acfLevels)
       {

          cout << "    " << cur << endl;
          
       }

       long double acfMean, acfLevelStdDev;

       cout << "ACF Level Mean = " << (acfMean = arithmeticMean(acfLevelValues)) << endl;
       cout << "ACF StandardDev= " << (acfLevelStdDev = standardDeviation(acfLevelValues)) << endl;

       set<size_t> timeIndices;

       long double minLevel = 1 - acfLevelStdDev;

       {
           
           
           
           CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, acfLevels)
           {
              if(cur.first >= minLevel)
              {
                  timeIndices.insert(cur.second);
              }
              
           }

           cout << "Main sequence maxima:" << endl;
           CXXTLS_FOREACH(size_t cur, timeIndices)
           {
              cout << "    " << cur << endl;
           }
           
           vector<size_t> timeDeltas;
           
           cout << "computing deltas in the time index series: " << endl;

           size_t largestDelta = 0;
           
           {
               set<size_t>::iterator first = timeIndices.begin(), last = timeIndices.end();
           
               while(first != last)
               {
                  size_t cur = *first++;
           
                  if(first != last)
                  {
                     long double delta = *first - cur;
           
                     timeDeltas.push_back(delta); 
           
                     if(delta > largestDelta)
                        largestDelta = delta;

                     cout << "    [" << *first << ", " << cur << "] = " << delta << endl;
                  }
           
           
               }
           }

           size_t meanTimeDelta = ceil(arithmeticMean(timeDeltas));
           size_t meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           
           cout << "Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;


           cout << "Removing large items" << endl;


           while(largestDelta > meanTimeStdDev)
           {
               // remove items out of range

               vector<size_t> tmp;

               largestDelta = 0;

               CXXTLS_FOREACH(size_t cur, timeDeltas)
               {
                   if(cur < meanTimeStdDev)
                   {
                      if(cur > largestDelta)
                         largestDelta = cur;

                      tmp.push_back(cur);
                   }

               }

               if(tmp.empty())
                  break;

               cout << "timeDeltas = " << timeDeltas << endl;
               cout << "tmp        = " << tmp        << endl;

               timeDeltas.swap(tmp);

               cout << "after swap timeDeltas = " << timeDeltas << endl;

               // recompute statistics

               meanTimeDelta = ceil(arithmeticMean(timeDeltas));
               meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           }


           cout << "Filtered Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Filtered Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;

           largestDelta = maximum(timeDeltas);

           cout << "Removing Time deltas less than " << largestDelta - meanTimeStdDev << endl;

           {
              vector<size_t> tmp;

              CXXTLS_FOREACH(size_t cur, timeDeltas)
              {
                 if(cur >= largestDelta - meanTimeStdDev)
                 {
                    tmp.push_back(cur);
                 }
              }

              if(tmp.size())
                 timeDeltas.swap(tmp);

           }

           cout << "Filtered Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Filtered Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;

           cout << "Final time deltas" << endl;

           CXXTLS_FOREACH(size_t cur, timeDeltas)
           {
              cout << "    " << cur << endl;
           }




       }  // main sequence

       {
           TimeSeriesAnalysis tsa(vi);
           

           cout << "levels of acf maxima: " << endl;
    
    
           CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, tsa.acfPeriods_)
           {
    
              cout << "    " << cur << endl;
              
           }

           cout << "ACF Level Mean = " << tsa.acfLevelMean_   << endl;
           cout << "ACF StandardDev= " << tsa.acfLevelStdDev_ << endl;


           cout << "Main sequence maxima:" << endl;

           CXXTLS_FOREACH( size_t cur, tsa.acfMainSequenceTimeIndices_)
           {
              cout << "    " << cur << endl;
           }




           cout << "TSA(VI) time period widths" << endl;


           CXXTLS_FOREACH(size_t cur, tsa.acfPeriodWidths_)
           {
              cout << "    " << cur << endl;
           }
       }


   }

   cout << "======================================================" << endl;
   cout << "Pittsburg hourly temp measurements (november 2013)  NOAA." << endl;
   {
       vector<int>   vi;

       static int table[]=
       {
#include "noaa_pittsburg.h"
       };

       CXXTLS_FOREACH(int cur, table)
          vi.push_back(cur);



       cout << "time series= " << vi << endl;
       cout << "mean of vi = " << arithmeticMean(vi) << endl;


       vector<int> maxima;

       vector<long double> acf;

       for(int i = 0; i < int(vi.size()-1); ++i)
       {
           long double value = autoCorrelationFunction(vi,i);

           acf.push_back(value);

           cout << "autoCorrelationFunction(" << i << ") = " << autoCorrelationFunction(vi,i) << endl;

       }

       cout << "acf mean = " << arithmeticMean(acf) << ", stdDev = " << standardDeviation(acf) << endl;
       cout << "acf min  = " << minimum(acf) << ", max = " << maximum(acf) << endl;

       bool goingDown = true;
       long double prev    = acf[0];
       size_t prevI   = 0;

       typedef map<long double, size_t> ACF_LEVEL_TYPE;
       typedef ACF_LEVEL_TYPE::value_type ACF_LEVEL_ENTRY;


       ACF_LEVEL_TYPE    acfLevels;
       vector<long double>    acfLevelValues;

       acfLevels.insert(ACF_LEVEL_ENTRY(1.0,0));

       for(size_t i = 1; i < acf.size(); ++i)
       {
          long double current = acf[i];

          if(goingDown)
          {
              if(current <= prev)
              {
                 prev = current;
                 continue;
              }
              else
                 goingDown = false; // ie we are now going up
          }
          else
          {
             // going up

             if(current >= prev)
             {
                prev = current;
                continue;
             }
             else
             {
                // prev is a local maximum
                goingDown = true;  // now we are falling
                prevI = i -1;
                cout << "Found a maximum at i = " << prevI << ", value = " << prev << endl;

                acfLevels.insert( ACF_LEVEL_ENTRY(prev,prevI) );
                acfLevelValues.push_back(prev);
             }

          }
       }

       cout << "levels of acf maxima: " << endl;


       CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, acfLevels)
       {

          cout << "    " << cur << endl;
          
       }

       long double acfMean, acfLevelStdDev;

       cout << "ACF Level Mean = " << (acfMean = arithmeticMean(acfLevelValues)) << endl;
       cout << "ACF StandardDev= " << (acfLevelStdDev = standardDeviation(acfLevelValues)) << endl;

       set<size_t> timeIndices;

       long double minLevel = 1 - acfLevelStdDev;

       {
           
           
           
           CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, acfLevels)
           {
              if(cur.first >= minLevel)
              {
                  timeIndices.insert(cur.second);
              }
              
           }

           cout << "Main sequence maxima:" << endl;
           CXXTLS_FOREACH(size_t cur, timeIndices)
           {
              cout << "    " << cur << endl;
           }
           
           vector<size_t> timeDeltas;
           
           cout << "computing deltas in the time index series: " << endl;

           size_t largestDelta = 0;
           
           {
               set<size_t>::iterator first = timeIndices.begin(), last = timeIndices.end();
           
               while(first != last)
               {
                  size_t cur = *first++;
           
                  if(first != last)
                  {
                     long double delta = *first - cur;
           
                     timeDeltas.push_back(delta); 
           
                     if(delta > largestDelta)
                        largestDelta = delta;

                     cout << "    [" << *first << ", " << cur << "] = " << delta << endl;
                  }
           
           
               }
           }

           size_t meanTimeDelta = ceil(arithmeticMean(timeDeltas));
           size_t meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           
           cout << "Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;


           cout << "Removing large items" << endl;


           while(largestDelta > meanTimeStdDev)
           {
               // remove items out of range

               vector<size_t> tmp;

               largestDelta = 0;

               CXXTLS_FOREACH(size_t cur, timeDeltas)
               {
                   if(cur < meanTimeStdDev)
                   {
                      if(cur > largestDelta)
                         largestDelta = cur;

                      tmp.push_back(cur);
                   }

               }

               if(tmp.empty())
                  break;

               cout << "timeDeltas = " << timeDeltas << endl;
               cout << "tmp        = " << tmp        << endl;

               timeDeltas.swap(tmp);

               cout << "after swap timeDeltas = " << timeDeltas << endl;

               // recompute statistics

               meanTimeDelta = ceil(arithmeticMean(timeDeltas));
               meanTimeStdDev = ceil(standardDeviation(timeDeltas));

           }


           cout << "Filtered Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Filtered Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;

           largestDelta = maximum(timeDeltas);

           cout << "Removing Time deltas less than " << largestDelta - meanTimeStdDev << endl;

           {
              vector<size_t> tmp;

              CXXTLS_FOREACH(size_t cur, timeDeltas)
              {
                 if(cur >= largestDelta - meanTimeStdDev)
                 {
                    tmp.push_back(cur);
                 }
              }

              if(tmp.size())
                 timeDeltas.swap(tmp);

           }

           cout << "Filtered Main Sequence Time Series width Delta mean = " << meanTimeDelta << endl;
           cout << "Filtered Main Sequence Time Series width Delta stddev = " << meanTimeStdDev << endl;

           cout << "Final time deltas" << endl;

           CXXTLS_FOREACH(size_t cur, timeDeltas)
           {
              cout << "    " << cur << endl;
           }




       }  // main sequence

       {
           TimeSeriesAnalysis tsa(vi);
           

           cout << "levels of acf maxima: " << endl;
    
    
           CXXTLS_FOREACH( ACF_LEVEL_ENTRY const &cur, tsa.acfPeriods_)
           {
    
              cout << "    " << cur << endl;
              
           }

           cout << "ACF Level Mean = " << tsa.acfLevelMean_   << endl;
           cout << "ACF StandardDev= " << tsa.acfLevelStdDev_ << endl;


           cout << "Main sequence maxima:" << endl;

           CXXTLS_FOREACH( size_t cur, tsa.acfMainSequenceTimeIndices_)
           {
              cout << "    " << cur << endl;
           }




           cout << "TSA(VI) time period widths" << endl;


           CXXTLS_FOREACH(size_t cur, tsa.acfPeriodWidths_)
           {
              cout << "    " << cur << endl;
           }
       }


   }

}
