
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.


//@File
//
///   This file defines a test for the Viewer class by creating a
///   trivial application that creates a single viewer.

#include <cxxtls/viewer.h>
#include <cxxtls/viewermanager.h>

using namespace std;
using namespace cxxtls;

class ExampleViewer
//
///   Trivial example of a viewer implementation.  This program prompts
///   the user with a single viewer, no other viewers can be created, and
///   the viewer doesn't support scrolling or cursor movements or data entry
///   it is just about the most trivial possible viewer.
//
: public Viewer
{
public:

  bool active_;

  ExampleViewer(ViewerManager*vm);

  typedef CursorWindow::viewport      viewport;    ///  simple typename alias
  typedef CursorWindow::input_event   input_event; ///  simple typename alias

  void operator() (viewport           * vp, int        cmd); ///  repaint handler
  bool handle_event(input_event const * e,  viewport * vp);  ///  key handler

  static std::string app_name;

  std::string const& application_name() const { return app_name; }

  std::string description() const { return "some description this is"; }

};

std::string ExampleViewer::app_name = "example";


typedef CursorWindow::viewport   viewport;
typedef CursorWindow::row_col    row_col;

ExampleViewer::
ExampleViewer(ViewerManager*vm)
: Viewer(vm),
  active_(false)
{
}



void
ExampleViewer::
operator() (viewport     * vp, int        cmd)
{
  if(cmd == CursorWindow::viewport::repaint_handler::activate)
    active_ = true;
  else
  if(cmd == CursorWindow::viewport::repaint_handler::deactivate)
    active_ = true;

  vp->set_curpos(0,0);
  vp->set_text_attribute(ViewerManager::normal_att);
  vp->fill_to_eos();   // empty the viewport

  int left = 0;  // offset from left side of viewport where drawing occurs

  CursorWindow::row_col origin = vp->origin();
  CursorWindow::row_col size   = vp->size();

  if(origin.col_ != 0)
  {
    left = 1;

    // leave room for column of vertical bars unless we are on the left edge
    // of the screen

    int i;

    // draw a left border (not needed in this example, but in the real world you
    // need to do this

    vp->set_text_attribute(ViewerManager::inactive_mark_att);

    for(i=0; i < size.row_; ++i)
    {
      vp->set_curpos(i, 0);
      vp->write('|');
    }

  }

  // all drawing should be biased by left (in column terms);

  // now handle the title:

  if(active_)
    vp->set_text_attribute(ViewerManager::active_title_att);
  else
    vp->set_text_attribute(ViewerManager::inactive_title_att);

  vp->set_curpos(0,left);
  vp->write("Title:  press q to quit");
  vp->fill_to_eol();

  // now handle the bottom of the viewer page

  vp->set_text_attribute(ViewerManager::normal_att);

  int row;

  for(row=1; row < size.row_; ++row)
  {
    vp->set_curpos(row, left);

    if(row == size.row_-1)
      vp->set_text_attribute(ViewerManager::bottom_att);  // special highlight at bottom line

    *vp << "This is row " << row;
  }

  vp->fill_to_eol();

  vp->set_curpos(0,left);



}

bool
ExampleViewer::
handle_event(input_event const * e,  viewport * vp)
{
  vp->set_curpos(1,0);
  vp->fill_to_eol();
  vp->set_curpos(1,0);

  *vp << CursorWindow::key_name(e->value_);

  if(e->value_ == 'q')
    return true;

  return false;
}


int main()
{
  ViewerManager vm;

  vm.open();

  vm.add( new ExampleViewer(&vm), "a viewer");

  vm.run();

}

