#ifdef  __linux__

#include <cxxtls/cmdterminalviewer.h>
#include <cxxtls/file.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pty.h>
#include <stdlib.h>
#include <unistd.h>

namespace cxxtls
{
    Viewer*
    CmdTerminalViewer::
    shellApp(ViewerManager*vm, std::string &fullname)
    {
        DumbTerminalViewer *vp =new CmdTerminalViewer(vm, "/bin/bash");

        vp->setTitle("/bin/bash");

        return vp;
    }


    CmdTerminalViewer::
    CmdTerminalViewer(ViewerManager *vm, std::string const &command)
    :  DumbTerminalViewer(vm, forkCommand(command))
    ,  waitingToDie_(false)
    {
        if(!ptyGood())
        {
            outputLine("Error opening pty");
        }
        else
        {
             // status is good

             waitThread_ = createThread(waitor, this, false);  // waitor will tell terminal that it is time for it to die

        }



    }

    CmdTerminalViewer::
    ~CmdTerminalViewer()
    {
       bool firstTime = true;

       for(int i = 0; i < 1000; ++i)  // at most 10 seconds hang
       {
          bool running = waitingToDie(); // if dead, not waiting to die

          if(!running)
             break;
          else
          if(firstTime)
          {
              firstTime = false;
              kill(childPid_, SIGHUP);   // terminate the process
          }

          usleep(10000);  // sleep 10 miliseconds

       }
    }

  
    void* 
    CmdTerminalViewer::
    waitor(void *p)
    {
        CmdTerminalViewer* me = reinterpret_cast<CmdTerminalViewer*>(p);

        int status=0;

        me->setWaitingToDie(true);

        // me->outputLine("waitor thread is alive");

        for(;;)
        {

            int rc = waitpid(me->childPid_, &status, 0);

            if(rc < 0)
            {
                char buffer[256];

                snprintf(buffer, sizeof(buffer), "Error:  unexpected error, %s, waiting for child process (%d)", strerror(errno), me->childPid_);

                me->outputLine(buffer);

                me->setWaitingToDie(false);

                return 0;  // let the user see the error message and close the window himself
            }

            if(WIFEXITED(status) || WIFSIGNALED(status) )
               break;
        }

        // me->outputLine("waitor:  child process is dead");

        me->setWaitingToDie(false);  // we are dying

        me->terminate();  // end the viewer 

        return 0;

    }

    int 
    CmdTerminalViewer::
    forkCommand(std::string const &commandValue)
    {
        std::string command;

        command.append(commandValue.begin(), commandValue.end());  // make a stack local copy of the string

        int masterFD=-1;

        int rc = forkpty(&masterFD, 0,0,0);

        if(rc == -1)
        {
           // error, could not fork
           return -1;
        }
        else
        if(rc == 0)
        {
            // success, we are in the child process, exec command

                setenv("TERM", "dumb", true);

                // rc = execl("/bin/bash", "/bin/bash", "-c", "TERM=dumb /bin/bash -i", (char*)0);

                rc = execl("/bin/bash", "/bin/bash", "-c", command.c_str(), (char*)0);

                if(rc < 0)
                {
                   int code =errno;

                   char buffer[256];

                   snprintf(buffer, sizeof(buffer), "Error %s:  ", strerror(code));

                   write(1, buffer, strlen(buffer));

                   write(1, command.c_str(), command.size());

                   write(1, "\n", 1);


                }


            // we should never get here

            exit(1);

        }

        // if we get here, rc contains the pid of the child process, and now we are in the parent

        childPid_ = rc;

        return masterFD;


    }


} // namespace cxxtls
#else
   int please_ignore_cmdterminalviewer_cxx=1;
#endif
