//
// Copyright 2010, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <cxxtls/csvviewer.h>
#include <cxxtls/file.h>
#include <cxxtls/excel.h>
#include <cxxtls/csv.h>
#include <cxxtls/viewermanager.h>
#include <iostream>
#include <fstream>
#include <cxxtls/strtool.h>
#include <cxxtls/file.h>
#include <portable_io.h>

using namespace std;

namespace cxxtls
{

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
CsvViewer::~CsvViewer()
{
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
CsvViewer::CsvViewer(ViewerManager *vm, string const &filename, int delimiter)
: TableEditor(vm, delimiter)
, filename_(filename)
{
    ifstream input(filename.c_str());

    if(!input.good())
    {

        CursorWindow::Dialog d("File does not exist.  Specify Rows and Columns");

        d += new CursorWindow::Dialog::String("rows",
                                              "Rows",
                                              "2",
                                              4
                                             );

        d += new CursorWindow::Dialog::String("cols",
                                              "Columns",
                                              "2",
                                              4
                                             );

        
        if(!d.popup( manager_->window() ))
        {

            int r=0;
            int c=0;

            string tmp = d.element_value("rows");

            sscanf(tmp.c_str(), "%d", &r);

            tmp = d.element_value("cols");

            sscanf(tmp.c_str(), "%d", &c);
            
            if(r < 1)
                r = 2;

            if(c < 1)
                c = 2;

            ofstream tmpfile(filename.c_str());

            if(!tmpfile.good())
            {
               fileError_="Can not create file ";
               fileError_ += filename;
               return;
            }

            // output header line

            for(int col = 0; col < c - 1 ; ++col)
                  tmpfile << decimal2excel(col+1) << ",";

            tmpfile << decimal2excel(c) << "\n";


            for(int row = 0; row < r; ++row)
            {
               for(int col = 0; col < c - 1 ; ++col)
                  tmpfile << ",";


               tmpfile << "\n";
            }

            input.close();

            input.open(filename.c_str());


        }
        else
        {
           
           fileError_="Can not open file ";
           fileError_+= filename;
           return;
       }
    }

    string line;

    bool firstLine=true;

    size_t rowCount=0;

    bool truncatedColumns = false; // if secondary lines have too many commas

    while(getline(input, line))
    {
        vector<string> fields;

        if(!line.empty())
        {
           if(line[ line.size() - 1 ] == '\r')
           {
              line.erase(line.size()-1);  // fix up dos to unix confusions.
           }
        }

        // StrTool::splitOnDelimiter(line, fields,StrTool::Is_Comma());

        csv::split(line, fields, ',');

        if(firstLine)
        {
            // if this is the first line, set up the column headers

            firstLine = false;

            for(size_t i = 0; i < fields.size(); ++i)
            {
                string tmp = decimal2excel(i+1);
                size_t width = fields[i].size();

                if(tmp.size() > width)
                   width = tmp.size();

                addColumnHeader(tmp, width, true);
            }

        }

        // on all lines, do the following
        
        {
            char buffer[40];

            sprintf(buffer, "%lu", (unsigned long)rows()+1);

            addRow(buffer,1,false);

            ++rowCount;

            for(size_t i = 0; i < fields.size(); ++i)
            {
               string const &cur = fields[i];


               if(cols_.size() > i)
               {
                  if(cols_[i].width_ < cur.size())
                     cols_[i].width_ = cur.size();
               }
               else
                  truncatedColumns = true;

               addColumnData(cur);
            }



        }

    }


    if(layout_.empty())
    {
       if(cols_.empty())
       {
          addColumnHeader("A", 3, true);
       }
       addRow("1", 1, false);

       for(size_t i = 0; i < cols_.size(); ++i)
          addColumnData("");
    }
    else
    if(cols_.empty())
    {
       addColumnHeader("A", 3, true);

       addColumnData("");
    }


    if(truncatedColumns)
    {

        CursorWindow::Message d("CSV File Format Mismatch");

        d += "";
        d += "Some lines of this file had a different number of fields  ";
        d += "than were found on row 1.";
        d += "";
        d += "Row 1 determines the column count";
        d += "";
        d += "Extra columns are simply ignored.  Abandon this session";
        d += "or risk file corruption.";
        d += "";
        d += "Press Enter to continue";
        d += "";
        
        d.popup( manager_->window() );
    }


}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
bool CsvViewer::okToQuit(CursorWindow::viewport *vp)
{
    if(is_dirty())
    {
      vp->beep();
      
      CursorWindow::Dialog d("Unsaved edit changes exist");
      d += new CursorWindow::Dialog::String("ok",
                                            "Quit Anyway (Yes or No)",
                                            "No",
                                            4
                                           );
      
      if(d.popup( manager_->window() ))
      {
        return false;
      }
      
      std::string ok_value = d.element_value("ok");
      
      if(ok_value.size() > 2)
      {
        ok_value.erase(2, ok_value.size());
      }

      // ok value is now only 2 characters long.

      if(   ok_value != "Ye" 
         && ok_value != "ye" 
         && ok_value != "YE"
        )
      {
        vp->beep();
        return false;
      }
      
    }
    return true;
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
bool CsvViewer::handle_event( CursorWindow::input_event const * e,
                              CursorWindow::viewport          * vp
                            )
{

  if(e->type_ == CursorWindow::input_event::ForceExitKey)
  {

    return true;
  }


    if(    e->type_ == CursorWindow::input_event::DataKey
       ||  e->type_ == CursorWindow::input_event::FunctionKey 
      )
    {
        if(keyKludge_)
        {
            // the ^X key was pressed earlier and now we are waiting for a second key in a sequence

            keyKludge_ = false;

            switch(e->value_)
            {
                case 0x03: // ^C quit editing
                   return true;

                case 0x13: // ^S         // save current file
                   saveTo(filename_);
                   return false;
                   

                case 0x0b: // ^K quit editing
                    if(okToQuit(vp))
                        return true;
                    return false;
                

                default: vp->beep(); break;
            }

            return false;
        }

        if(e->type_ == 2 && e->value_ == CursorWindow::key_f4)
        {
           // superseed the create column behavior.


             CursorWindow::Selection d("Before or after the current column?");
         
             d += "Before";
             d += "After";


             if(d.popup(manager_->window()))
               return false;
         

           CursorInfo tmp = cursorInfo();

           size_t width = cols_[tmp.col_].width_;


           if(d.selection_ == "After")
               ++tmp.col_;

           insertColumn(tmp.col_, decimal2excel(tmp.col_+1), width, "");

           pushUndoColCreate(tmp.col_, true);

           for(size_t i = 0; i < cols_.size(); ++i)
           {
               cols_[i].title_ = decimal2excel(i+1);
           }
                     
           (*this)(vp, CursorWindow::viewport::repaint_handler::activate);
            setCursorInfo(vp, tmp.row_, tmp.col_, true);



           return false;
        }

        if(e->type_ == 2 && e->value_ == CursorWindow::key_f6)
        {

             if(cols_.size() > 1 && layout_.size() >= 1)
             {
                 // there is at least 1 column

                 CursorInfo tmp = cursorInfo();

                 if(    tmp.row_ < layout_.size()
                    &&  !layout_[tmp.row_].empty()
                    &&  tmp.col_ < layout_[tmp.row_].size()
                   )
                 {
                     // we are sitting on a valid column

                     pushUndoColDelete(tmp.col_, true);

                     deleteColumn(tmp.col_); 


                     if(tmp.col_ >= layout_[tmp.row_].size())
                     {
                        if(layout_[tmp.row_].empty())
                           tmp.col_ = 0;
                        else
                           --tmp.col_;
                     }
             
                     for(size_t i = 0; i < cols_.size(); ++i)
                     {
                         cols_[i].title_ = decimal2excel(i+1);
                     }
                     
        
                     makeVisible(vp, tmp.row_, tmp.col_, true);
                     (*this)(vp, CursorWindow::viewport::repaint_handler::activate);
                 }
             }
            return false;
        }

    }

   // by default, act like a table viewer -- which has different arrow key behavior
   // than is inforced here

   if(TableEditor::handle_event(e,vp))
   {
       // the table editor or table viewer, one, has told us to quit.
       //
       // Those programs are designed to enable the ScriptTableViewer to allow
       // the script to interpret what the key means.
       //
       // Here, there is no one else to interpret -- we have to check to see if the
       // the key that caused the table editor to return true really should cause US
       // to return true.

       switch(e->value_)
       {
           case 3:  // ^C
           case 0x0b:  // ^K
           case CursorWindow::key_f2:  // F2, switch windows
              break;

           default:
               return false;  // very few keys should actually cause us to really terminate
               
       }

       if(okToQuit(vp))
             return true;
       return false;
   }


   return false;

}  // handle_event()


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
void CsvViewer::setCursorInfo(CursorWindow::viewport *vp,
                              size_t worldRow, 
                              size_t worldCol, 
                              bool displayed
                             )
{
    terminateInput();

    TableViewer::setCursorInfo(vp, worldRow, worldCol, displayed);
}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

void CsvViewer::saveTo(std::string const &filename)
    {
       set_data_dirty(false);

       string delimString;  delimString += (char)delimiter_;


       FileName tmp(filename);

       tmp.remove();  // delete the file before writing on it.

       ofstream output(filename.c_str());

       fileError_.resize(0);

       if(!output.good())
       {
          CursorWindow::Message d("Error opening file for write ");
        
          d += "";
          d += string("File name: ") + filename;
          d += "";
          d += "Press Enter to continue";
        
          d.popup(manager_->window());
   
          fileError_ = string("Error openining input file: ") + filename;

          return;
          
       }


       for(size_t row = 0; row < layout_.size(); ++row)
       {
          vector<CellInfo> const &curRow = layout_[row];

          vector<CellInfo>::const_iterator first = curRow.begin(),
                                           last  = curRow.end();

          while(first != last)
          {
              CellInfo const &curCell = *first++;
              
              if(!curCell.text_.empty())
                  output << csv::decorate(curCell.text_[0]); // handle internal quotes in the string

              if(first != last)
                  output << delimString;

          }

          output << '\n';

       }

       if(!output.good())
       {
          fileError_ = string("Error writing to: ") + filename;
       }


    }

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
std::string const &CsvViewer::application_name() const
{
   static string rv("CsvViewer");

   rv = filename_;

   return rv;
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
std::string CsvViewer::description() const
{
   return "csv and tsv file editor";
}
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
Viewer* CsvViewer::app(ViewerManager*vm, std::string &fullname)
{
  if(fullname.size() == 0)
  {
    // do not use the viewer selector from here because it can not repaint the screen
    // in this context

    CursorWindow::Dialog d("File name");

    d += new CursorWindow::Dialog::String("name",
                                          "path",
                                          "*",
                                          40
                                         );
                                
    if(d.popup(vm->window()))
      return 0;

    fullname = d.element_value("name");

    if(fullname == "")
      return 0;


    FileName tmp(fullname);
    tmp.convert_to_absolute_path();

    fullname = tmp;

  }

  int delimiter=',';

  size_t dotPos = fullname.find_last_of('.');

  if(dotPos < fullname.size())
  {
    ++dotPos;  // get to 't' or 'c'

    if( dotPos < fullname.size())
    {
        if(fullname[dotPos] == 't' || fullname[dotPos] == 'T')
           delimiter=0x09; // tab  
    }
  }

  return new CsvViewer(vm,fullname, delimiter);
}

void
CsvViewer::
help()
{

  std::auto_ptr< std::list< std::string > >
    viewer_help( new std::list< std::string > );

  viewer_help->push_back("");
  viewer_help->push_back("CSV FILE VIEWER");
  viewer_help->push_back("");
  viewer_help->push_back("Warning:  work in progress!");
  viewer_help->push_back("");
  viewer_help->push_back("Note that this is mainly a viewer with a few editing capabilities.");
  viewer_help->push_back("");
  viewer_help->push_back("  Movement keys");
  viewer_help->push_back("    up arrow    -- move up one cell");
  viewer_help->push_back("    down arrow  -- move down one cell");
  viewer_help->push_back("    left arrow  -- move left 1 character or cell");
  viewer_help->push_back("                   if at the beginning of the cell");
  viewer_help->push_back("    right arrow -- move right 1 character or cell");
  viewer_help->push_back("                   if at the end of text in the cell");
  viewer_help->push_back("    Home        -- move to column A on the current row.");
  viewer_help->push_back("    End         -- move to the last column on the current row.");
  viewer_help->push_back("    tab         -- move right one cell");
  viewer_help->push_back("    shift-tab   -- move left one cell");
  viewer_help->push_back("    ^A          -- move to left of the text in the cell");
  viewer_help->push_back("    ^E          -- move to the right of the text in the cell");
  viewer_help->push_back("    ^Q          -- move left 1 whole page");
  viewer_help->push_back("    ^P          -- move right 1 whole page");
  viewer_help->push_back("");
  viewer_help->push_back("  Editing keys");
  viewer_help->push_back("    Delete      -- deletes the current character");
  viewer_help->push_back("    Backspace   -- deletes the previous character");
  viewer_help->push_back("    F3          -- delete marked block");
  viewer_help->push_back("    F4          -- adds a column");
  viewer_help->push_back("    F5          -- toggle inclusion of current row in marked region");
  viewer_help->push_back("    F6          -- deletes a column");
  viewer_help->push_back("    F7          -- inserts paste buffer rows above the current line");
  viewer_help->push_back("    F8          -- deletes the current row");
  viewer_help->push_back("    F9          -- inserts a new row above");
  viewer_help->push_back("    F11         -- copies the marked block to the paste buffer");
  viewer_help->push_back("    F12         -- convert title line to row");
  viewer_help->push_back("    ctrl shift _-- undo (control shift underscore)");
  viewer_help->push_back("    ^Z          -- undo");
  viewer_help->push_back("");
  viewer_help->push_back("  Other keys");
  viewer_help->push_back("");
  viewer_help->push_back("    printable characters get inserted into the cell");
  viewer_help->push_back("");

  manager_->help_helper(*viewer_help);
}

} // namespace cxxtls
