//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//                            
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file
/// Implements members of the CursesInterface namespace.
///
/// Note that this file is mainly just a trivial wrapper around the source code
/// found in header files.  There  are two:  @ref unix_curses.h and @ref windows_curses.h
///
/// These files are named .h but they are really full blown C++ source code.  Only
/// one set of which should actually be compiled -- either the unix or the windows
/// implementation of curses. 
///
/// The windows implementation uses an old copy of pdcurses pulled off the net
/// a long time ago and which has been tweaked in a few places for my needs.
/// It has it's own copyright by the way.

#include <cxxtls/cursesinterface.h>

//
// There are 3 implementations of the curses interface
//   1.  the unix normal implementation (truly portable)
//   2.  the microsoft windows implementation (goofy but workable)
//   3.  the test implementation (an ascii only version for diagnostic tests)
//       which can be found in ../tests/curses_stub.c
// The version you get is based on the compiler you are using:  if you are using
// the microsoft compiler, the assumption is that you want the windows version.
//
// To use the diagnostic version, manually build the ../tests/curses_stub.$O file
// and force it onto the link command line rather than relying on the $(TOOLS)
// library to supply the one selected by this file.
//

#ifdef _MSC_VER
# include <windows_curses.h>
#else
# include <unix_curses.h>
#endif

namespace cxxtls
{

namespace CursesInterface
{
  bool winch_occurred = false;
  bool resize_requested = false;
}

} // namespace cxxtls
