#include <cxxtls/deltaviewer.h>
#include "portable_io.h"             // for snprintf
//===========================================================================
//===========================================================================
//===========================================================================

namespace
{
    const int TITLE_HEIGHT= 1;
    const int EXPR_HEIGHT = 1;
    const int VALUE_HEIGHT= 1;


    int availableRows(cxxtls::CursorWindow::viewport *vp)  { return vp->size().row_ - TITLE_HEIGHT - VALUE_HEIGHT - EXPR_HEIGHT; }
};

namespace cxxtls
{

void
DeltaViewer::
paintCurrentExpression(CursorWindow::viewport *     vp)
{
    vp->set_curpos(vp->size().row_ - EXPR_HEIGHT - VALUE_HEIGHT, 0);

    NameValuePair const *p=0;

    if(cursorRow_ < int(data_.size()))
    {    
        std::list<NameValuePair>::iterator iter =  data_.begin();

        int countDown = cursorRow_;

        while(countDown--)
           ++iter;

        p = &*iter;

    }

    vp->write("|EXPR :");

    if(p)
        vp->write(p->name_);

    vp->fill_to_eol();

    vp->set_text_attribute(CursorWindow::bottom_att);

    vp->set_curpos(vp->size().row_ - VALUE_HEIGHT, 0);
    vp->write("|VALUE:");

    CursorWindow::row_col curpos = vp->curpos();

    if(p)
        vp->write(p->value_);


    vp->fill_to_eol();

    vp->set_curpos(curpos);

}



void 
DeltaViewer::
paintPage(CursorWindow::viewport *     vp)
{

    paintTitle(vp);

    CursorWindow::row_col size = vp->size();

    int remainingRows = availableRows(vp);

    int row = TITLE_HEIGHT;

    vp->set_text_attribute(CursorWindow::normal_att);

    int rowOffset = topDisplayedRow_; // offset from top of data_

    int NonCursorNonModified = CursorWindow::normal_att;
    int NonCursorModified    = CursorWindow::bottom_att;

    int CursorNonModified    = CursorWindow::active_mark_att;
    int CursorModified       = CursorWindow::inactive_mark_att;


    std::list<NameValuePair>::iterator iter = data_.begin();

    int countDown = rowOffset;  // offset to start row to be displayed

    while(countDown--)
        ++iter;

    // iter now points to data_ location corresponding to row_offset


    while(iter != data_.end())
    {                        
        if(remainingRows < 0)
           break;

        NameValuePair const &cur = *iter;

        int lineAtt  = 0;
        int valueAtt = 0;

        if(rowOffset == cursorRow_)
        {
           // cursor row

           lineAtt = CursorNonModified;
           valueAtt= cur.changed_ ? CursorModified : CursorNonModified;

        }
        else
        {
           // non-cursor rows

           lineAtt = NonCursorNonModified;
           valueAtt= cur.changed_ ? NonCursorModified : NonCursorNonModified;

        }

        vp->set_text_attribute(CursorWindow::normal_att);
        vp->set_curpos(row,0);
        vp->write("|");                    // this viewer is always meant to be used in an horizontal split,
                                           // so the normal test for "am I split horizontally" makes no sense here -- just
                                           // pretend we are are under all circumstances.


        vp->set_text_attribute(lineAtt);
        vp->write(cur.name_);
        vp->fill_to_eol();

        vp->set_curpos(row, size.col_ / 3);
        vp->set_text_attribute(valueAtt);
        vp->write(cur.value_);
        
        vp->set_text_attribute(lineAtt);
        vp->fill_to_eol();

        ++row;
        --remainingRows;
        ++rowOffset;
        ++iter;
        
    }

    vp->set_text_attribute(CursorWindow::normal_att);

    while(remainingRows--)
    {
       vp->set_curpos(row++, 0);
       vp->write("|");
       vp->fill_to_eol();
    }

    paintCurrentExpression(vp);
}

//===========================================================================
//===========================================================================
//===========================================================================

void 
DeltaViewer::
paintTitle(CursorWindow::viewport *     vp)
{
    vp->set_curpos(0,0);
    vp->set_text_attribute(CursorWindow::normal_att);
    vp->write("|");                                           // always assume we are in split screen mode
    vp->set_text_attribute(CursorWindow::active_title_att);
    vp->write(title_);
    vp->fill_to_eol();

}



//===========================================================================
//===========================================================================
//===========================================================================

void 
DeltaViewer::
operator() ( CursorWindow::viewport *     vp,
             int                          cmd
           )
{
    displayMode_ = cmd;

    // paint the entire screen on command

    paintPage(vp);

}


//===========================================================================
//===========================================================================
//===========================================================================

Viewer* DeltaViewer::app(ViewerManager*vm, std::string &title)
{
    // create a new viewer of this type and return it

    return new DeltaViewer(vm, title);
}

//===========================================================================
//===========================================================================
//===========================================================================

std::string DeltaViewer::app_name("DeltaViewer");

//===========================================================================
//===========================================================================
//===========================================================================

std::string DeltaViewer::description() const { return "Process Watch Variables"; }

//===========================================================================
//===========================================================================
//===========================================================================

void DeltaViewer::help() 
{
}

//===========================================================================
//===========================================================================
//===========================================================================

bool DeltaViewer::handle_event( CursorWindow::input_event const * event,
                                CursorWindow::viewport          * viewport
                              )
{

    if(event->type_ == CursorWindow::input_event::ForceExitKey)
       return true; // someone did ^X^C in another window

    int finalRow = int(data_.size()) - 1;

    int pageHeight = availableRows(viewport);

    int rowOffset = cursorRow_ - topDisplayedRow_;


    if(event->type_ == CursorWindow::input_event::FunctionKey)
    {
        int edit_func = (*vm_->window()).func[event->value_];

        switch(edit_func)
        {
             case CursorWindow::func_enter:

                 return false; // for the moment.  Probably needs to be tied to editing capabilities.    
                 break;


             case CursorWindow::func_top:
                     if(!data_.empty())
                     {
                         topDisplayedRow_ = 0;
                         cursorRow_       = 0;
                         paintPage(viewport);
                     }

                 break;

             case CursorWindow::func_bottom:
                     if(!data_.empty())
                     {
                         topDisplayedRow_ = finalRow;
                         cursorRow_       = finalRow;
                         paintPage(viewport);
                     }

                 break;



             case CursorWindow::func_up:
                     // move cursor up

                     if(cursorRow_ > topDisplayedRow_)
                     {
                         --cursorRow_;
                         paintPage(viewport);
                     }
                     else
                     {
                         if(topDisplayedRow_)
                         {
                            --topDisplayedRow_;
                            cursorRow_ = topDisplayedRow_;
                            paintPage(viewport);
                         }
                         else
                         {
                            viewport->beep();

                         }
                     }

                     return false;

                 break;

             case CursorWindow::func_down:
                     // move cursor down

                     if(rowOffset + 1 >= pageHeight)
                     {
                         // need to scroll down 1 line to see the one we
                         // are about to uncover, and there is room to do
                         // so

                         if(    (topDisplayedRow_ < finalRow)
                            &&  (cursorRow_ < finalRow)
                           )
                         {

                             ++cursorRow_;
                             ++topDisplayedRow_;
                             paintPage(viewport);
                         }
                         else
                         if(topDisplayedRow_ < finalRow)
                         {
                             ++topDisplayedRow_;
                             paintPage(viewport);
                         }
                         else
                            viewport->beep();

                     }
                     else
                     if(cursorRow_ < finalRow)
                     {
                        // room to move down on current page
                        ++cursorRow_;
                        paintPage(viewport);
                     }
                     else
                     if(topDisplayedRow_ < finalRow)
                     {
                        ++topDisplayedRow_;
                        paintPage(viewport);
                     }
                     else
                         viewport->beep();



                     return false;

                 break;

             case CursorWindow::func_left:
                     // when editing the value, move the cursor left, otherwise ?
                 break;

             case CursorWindow::func_right:
                     // when editing the value, move the cursor right, otherwise ?
                 break;

             case CursorWindow::func_prior:
                     // move cursor up one page

                     if(topDisplayedRow_ >= pageHeight)
                     {
                         topDisplayedRow_ -= pageHeight;
                         cursorRow_       -= pageHeight;
                     }
                     else
                     if(    (topDisplayedRow_ < pageHeight)
                        ||  (topDisplayedRow_ > finalRow)
                       )
                     {
                         topDisplayedRow_ = 0;
                         cursorRow_       = 0;
                     }
                     else
                         viewport->beep();
                     

                     paintPage(viewport);

                     return false;

                 break;

             case CursorWindow::func_next:
                     // move cursor up one page

                     if(topDisplayedRow_ <= finalRow - pageHeight)
                     {
                         topDisplayedRow_ += pageHeight;
                         cursorRow_       += pageHeight;

                     }
                     else
                     if(topDisplayedRow_ < finalRow)
                     {
                         topDisplayedRow_ = cursorRow_ = finalRow;
                     }
                     else
                         viewport->beep();
                     

                     paintPage(viewport);

                     return false;

                 break;


             default:

                 if(event->value_ == ('K' - '@'))
                 {
                    return true; // die
                 }
                 else
                 {
                     viewport->beep();     // ignore unknown commands
                     return false;
                 }
                 break;
        }

    }
    else
    if(event->type_ == CursorWindow::input_event::DataKey)
    {
       // eventually,  use non-function keys to change the values in the list -- informing the owner of this list as we go.
    }
    else
    if(event->type_ == CursorWindow::input_event::ForceExitKey)
    {
       return true;
    }
    else
    {
       viewport->beep();
       return false;
    }

    if(event->value_ == 'a')
    {
         update("fred", "12345");
         update("bill", "1");
         update("sue",  "12345");

         for(int i = 0; i < 50; ++i)
         {
             char buffer[40];

             snprintf(buffer, 40, "v%d", i);

             update(buffer, buffer);
         }

         paintPage(viewport);
         return false;
    }




    if(event->value_ == ' ')
    {
        if(!data_.empty())
        {
            data_.front().changed_ ^= true; 

            paintPage(viewport);

            return false;
        }
    }

    if(event->value_ == 'u')
    {
         std::list<NameValuePair>  newData;

         newData.push_back(NameValuePair("sue", "14"));
         newData.push_back(NameValuePair("hank", "fred"));

         replace(newData);

         paintPage(viewport);

         return false;
    }


    return false;   // if we get here, don't abandon the viewer session
}

static 
DeltaViewer::NameValuePair const* 
findInList(std::list<DeltaViewer::NameValuePair> const &data_, std::string const &name)   // return 0 if not found
    {
        CXXTLS_FOREACH(DeltaViewer::NameValuePair const & i, data_)
        {
            if(name == i.name_)
               return &i;
        }

        return 0;
    }


void DeltaViewer::replace(std::list<DeltaViewer::NameValuePair> const &data)
{
    cursorRow_ = topDisplayedRow_ = 0;

    CXXTLS_FOREACH(NameValuePair &cur, data_)  { cur.changed_ = false; }  // clear the changed flag on existing data

    CXXTLS_FOREACH(NameValuePair const &cur, data)
    {
         // set the changed flag on added or modified data

         NameValuePair  *match = find(cur.name_);
         
         if(match)
         {
             match->update(cur.value_);
         }
         else
         {
             data_.push_back(cur);  // changed_ flag set when cur was constructed
         }

    }

    // adds and updates are done.  check for deletes

    std::list<NameValuePair>::iterator f = data_.begin(), l = data_.end();
    while( f != l)
    {
        std::list<NameValuePair>::iterator old = f++;

        if(!findInList(data, old->name_))
           data_.erase(old); 

        // f is already increment past the victim we just deleted, so it is safe

    }



}

} // namespace deltaviewer
