#ifndef __cmdterminalviewer__h_included
#define __cmdterminalviewer__h_included

///@file
///   This file declares a viewer class, CmdTerminalViewer.

#include <cxxtls/dumbterminal.h>
#include <cxxtls/controlledprocess.h>

namespace cxxtls
{
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
class CmdTerminalViewer
:  public DumbTerminalViewer
    ///  A CmdTerminalViewer allows user interaction with programs running
    ///  in what they perceive to be a dumb terminal.
    ///
    ///  The program can read/write to the terminal (as stdin, stdout, and stderr)
    ///  and outputs will be displayed as if they were recorded in a scrollable
    ///  log file.  Inputs come from the user via an "input area" at the bottom
    ///  of the window.
    ///
    ///  Note that this kind of viewer is not good for editors or other
    ///  highly interactive programs but should be good for shells and other
    ///  programs that just dump data to standard out.  
    ///
    ///  Note that terminal is DUMB meaning no cursor positioning.  Many editor 
    ///  programs will not function in this environment at all.
    ///
    ///  Warning:  Do not use command line file redirection in the constructor's
    ///  command parameter -- it will be superseeded by the viewer and may cause
    ///  program start failure.
    ///
{
    ThreadId                          waitThread_;  // just waits for process to die then signals dumb terminal that it is time to die

    int forkCommand(std::string const &command);

public:

    CmdTerminalViewer(ViewerManager *vm, std::string const &command);

    ~CmdTerminalViewer();  // kill waitThread_ and wait for it to die

    static Viewer* shellApp(ViewerManager*vm, std::string &fullname);
         ///<  create a new dumb terminal app


private:

    static void* waitor(void*);   // the thread that waits for 

    Mutex terminateMutex_;
    bool  waitingToDie_;

    void setWaitingToDie(bool b)
    {
       ScopedMutexLock lock(terminateMutex_);
       waitingToDie_ = b;
    }
    
    bool waitingToDie()
    {
       ScopedMutexLock lock(terminateMutex_);
       return waitingToDie_;
    }

    int childPid_;


};   // class CmdTerminalViewer


}  // namespace cxxtls

#endif // __cmdterminalviewer__h_included
