#ifndef textviewer_included_h
#define textviewer_included_h
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file
/// Defines class TextViewer.


#include <cxxtls/viewer.h>
#include <cxxtls/file.h>

namespace cxxtls
{


class TextViewer
: public Viewer
///
/// A TextViewer is a Viewer meant for displaying but not editing
///  ascii text files.  It presents the user with the text of the named
///  file and allows for scrolling (vertically and horizontally) as well as
///  marking lines with the file for copying/pasting.
///
{
private:

  void ensure_cursor_visible(viewport*);
    ///<
    /// force any screen scrolling needed to make the cursor visible


public:

  struct impl;                     ///< implementation data not needing to be visible to users.

  std::auto_ptr<impl> impl_; ///<  Pointer to the implementation data

  TextViewer(ViewerManager*vm, FileName f);
 ~TextViewer();

  virtual void operator() ( CursorWindow::viewport *     vp,
			    int                          cmd
			  );
     ///< Handle screen repaint commands

  virtual bool handle_event( CursorWindow::input_event const * e,
			     CursorWindow::viewport          * vp
			   );
     ///< Handle input events.

  virtual void help();  ///< Present the user with help information that is
			///  specific to this viewer's operation and contains
			///  information about system keep bindings, etc.

  static Viewer* app(ViewerManager*vm, std::string &fullname);
    ///< Application "hook" function for the DirectoryViewer class.
    ///  Use it as a parameter to add_app, like this:
    ///  @code
    ///   vm.add_app(DirectoryViewer::app, "DirectoryViewer");
    ///  @endcode

  static std::string app_name;
    ///<
    ///  The name to prompt the user with in the ViewerManager::appChooser
    ///  menu.

  std::string const& application_name() const { return app_name; }
    ///<
    /// Return the name of the application to use when describing this
    /// viewer in the ViewerManager::appChooser or in the
    /// ViewerManager::ViewerSelector.

  std::string description() const;
    ///<
    /// Returns a description of the file being view -- its name only

};
} // namespace cxxtls
#endif
